# -*- coding: utf-8 -*-

""" OpenStreetMap related tools

Among available tools:
- Use Overpass API to query OSM database and convert JSON
response to geopandas dataframe (thanks to https://github.com/yannforget/OSMxtract for inspiration !)
"""
import geojson
import geopandas as gpd

from gistools import GEOMETRY_CLASS
from gistools.exceptions import QlQueryError
from gistools.geometry import merge
from osmnx import gdf_from_place, get_polygons_coordinates, overpass_request
from osmnx.settings import default_crs
from shapely.geometry import LineString, Point
from utils.check import check_string

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2019, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


def _to_point_features(json):
    """ Read json response and extract point geometries

    :param json: JSON response from overpass API
    :return: GeoJSON FeatureCollection
    """
    features = []
    elements = [e for e in json['elements'] if e['type'] == 'node']
    for elem in elements:
        coords = [elem['lon'], elem['lat']]
        features.append(geojson.Feature(id=elem['id'], geometry=Point(coords), properties=_feature_tags(elem)))

    return geojson.FeatureCollection(features)


def _to_features(json, geometry_type):
    """ Read json response and extract (multi)linestring/polygon geometries

    :param json: json response
    :param geometry_type: {'linestring', 'polygon'}
    :return:
    """
    features = []

    if geometry_type == 'linestring':
        elements = [e for e in json['elements'] if e['type'] == 'way' or e['type'] == 'relation']
    else:
        elements = [e for e in json['elements'] if e['type'] == 'way' or (e['type'] == 'relation' and e['tags']['type']
                                                                          in ('multipolygon', 'boundary'))]

    for elem in elements:

        if elem['type'] == 'way':
            coords = [[node['lon'], node['lat']] for node in elem['geometry']]
            try:
                geom = GEOMETRY_CLASS[geometry_type][0](coords)  # LineString, Polygon
            except ValueError:
                pass
            else:
                features.append(geojson.Feature(id=elem['id'], geometry=geom, properties=_feature_tags(elem)))

        elif elem['type'] == 'relation':
            collection = []
            for member in elem['members']:
                if member['type'] == 'way':
                    member_coords = [(node['lon'], node['lat']) for node in member['geometry']]
                    collection.append(LineString(member_coords))
            geom_collection = merge(collection)

            if geom_collection:
                try:
                    geom = GEOMETRY_CLASS[geometry_type][1]([GEOMETRY_CLASS[geometry_type][0](line) for line in
                                                             geom_collection])  # MultiLineString, MultiPolygon
                except ValueError:
                    pass
                else:
                    features.append(geojson.Feature(id=elem['id'], geometry=geom, properties=_feature_tags(elem)))

    return geojson.FeatureCollection(features)


def _feature_tags(json_element):
    """ Update feature tags to set OSM ID and type

    :param json_element: 'elements' feature in JSON dict response
    :return:
    """
    if 'id' not in json_element['tags'].keys():
        tags = dict(osm_id=json_element['id'], **json_element['tags'])
    else:
        tags = json_element['tags']

    # Add osm type to attributes
    tags.update(osm_type=json_element['type'])

    return tags


def download_osm_features(place, osm_type, tag, values=None, by_poly=True, timeout=180):
    """ Download OSM features within given place

    :param place: single place name query (e.g: "London", "Bonn", etc.)
    :param osm_type: OSM geometry type str ('node', 'way', 'relation')
    :param tag: OSM tag to query
    :param values: str/list of possible values for the provided OSM tag
    :param by_poly: if True, retrieve features within polygon's list of coordinates, otherwise use bounds
    :param timeout:
    :return:
    """
    gdf_geometry = gdf_from_place(place)

    try:
        geometry = gdf_geometry.geometry[0]
    except AttributeError:  # Empty GeoDataFrame
        return None

    responses = []

    if by_poly:
        polygon_coord_strs = get_polygons_coordinates(geometry)
        for poly_coord_str in polygon_coord_strs:
            query = ql_query(osm_type, tag, values, polygon_coord=poly_coord_str, timeout=timeout)
            responses.append(overpass_request(data={'data': query}, timeout=180))
    else:
        query = ql_query(osm_type, tag, values, bounds=geometry.bounds, timeout=timeout)
        responses.append(overpass_request(data={'data': query}, timeout=180))

    return responses


def json_to_geodataframe(response, geometry_type):
    """ Convert JSON responses to

    :param response: json response
    :param geometry_type: type of geometry to extract ('point', 'linestring', 'polygon', 'multipolygon')
    :return:
    """
    geometry_type = check_string(geometry_type, ('point', 'linestring', 'polygon'))

    if geometry_type == 'point':
        return gpd.GeoDataFrame.from_features(_to_point_features(response), crs=default_crs)
    else:
        return gpd.GeoDataFrame.from_features(_to_features(response, geometry_type), crs=default_crs)


def ql_query(osm_type, tag, values=None, bounds=None, polygon_coord=None, timeout=180):
    """ QL query (thanks to https://github.com/yannforget/OSMxtract for inspiration !)

    :param osm_type: OSM geometry type str {'node', 'way', 'relation', 'nwr'}
    :param tag: OSM tag to query
    :param values: str/list of possible values for the provided OSM tag
    :param bounds: geometry bounds
    :param polygon_coord: location's polygon list of coordinates
    :param timeout:
    :return:
    """
    osm_type = check_string(osm_type, ('node', 'way', 'relation', 'nwr'))

    if isinstance(values, str):
        values = [values]

    if bounds and not polygon_coord:
        west, south, east, north = bounds
        boundary = f'({south:.6f},{west:.6f},{north:.6f},{east:.6f})'
    elif polygon_coord and not bounds:
        boundary = f'(poly:"{polygon_coord}")'
    else:
        raise QlQueryError("Must define either geometry bounds or polygon coordinates")

    if values:
        if len(values) > 1:
            tags = f'["{ tag }"~"{ "|".join(values) }"]'
        else:
            tags = f'["{ tag }"="{ values[0] }"]'
    else:
        tags = f'["{tag}"]'

    return f'[out:json][timeout:{timeout}];{osm_type}{tags}{boundary};out geom;'



# -*- coding: utf-8 -*-

""" Raster map classes and methods

Toolset for working with static raster/array maps,
defined as matrices of cell values lying on a
geo-referenced grid
"""
from rasterio import open as rasterio_open
from rasterio.merge import merge as rasterio_merge
from urllib.error import URLError
from urllib.request import urlretrieve
from zipfile import ZipFile

import pyproj
import warnings
import os
import tempfile
from functools import wraps
import copy

import numpy as np
from osgeo import gdal, ogr
from matplotlib import pyplot as plt
from utils.check.type import check_type, collection_type_assert, type_assert, isfile
from utils.check.descriptor import protected_property
from utils.check.value import check_string

from gistools.coordinates import GeoGrid
from gistools.conversion import raster_to_array, array_to_raster
from gistools.exceptions import RasterMapError, GeoGridError, DigitalElevationModelError
from gistools.files import RasterTempFile, ShapeTempFile
from gistools.layer import PolygonLayer, check_proj
from gistools.projections import proj4_from_raster, is_equal, proj4_from, wkt_from, srs_from, ellipsoid_from
from gistools.surface import compute_surface

__all__ = ["RasterMap", "DigitalElevationModel"]
__version__ = '0.1'
__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2018, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'

gdal.UseExceptions()


CGIAR_URL = "http://srtm.csi.cgiar.org/wp-content/uploads/files/srtm_5x5/TIFF"
DEFAULT_OUTPUT = os.path.join(tempfile.gettempdir(), "out.tif")


# Decorator for returning new instance of RasterMap
def return_new_instance(method):
    @wraps(method)
    def _return_new_instance(self, *args, **kwargs):
        output = method(self, *args, **kwargs)
        if isinstance(output, np.ndarray):
            new_self = self.__class__(output, self.geo_grid, no_data_value=self.no_data_value, crs=self.crs)
        elif isinstance(output, tuple):
            new_self = self.__class__(output[0], output[1], no_data_value=self.no_data_value, crs=self.crs)  # Return
            # array AND GeoGrid
        else:
            new_self = output
        return new_self
    return _return_new_instance


def gdal_decorator(no_data_value=None):
    def decorate(method):
        no_data = no_data_value

        @wraps(method)
        def _gdal_decorator(self, *args, **kwargs):
            with RasterTempFile() as out_raster:
                method(self, out_raster, *args, **kwargs)
                if no_data is None:
                    return self.__class__(out_raster, no_data_value=self.no_data_value)
                else:
                    return self.__class__(out_raster, no_data_value=no_data)
        return _gdal_decorator
    return decorate


class GdalOpen:

    def __init__(self, raster_file):
        self.dataset = gdal.Open(raster_file)

    def __enter__(self):
        return self.dataset

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.dataset

# TODO: raster map with band number


class RasterMap:
    """ RasterMap base class

    A RasterMap is a numpy array corresponding to a geo raster
    When the raster is stored using numpy array, and when it is
    possible, methods rely on numpy/rasterio packages, otherwise
    gdal is used
    """

    # Mapping numpy types to gdal types (thanks to
    # https://gist.github.com/CMCDragonkai/ac6289fa84bcc8888035744d7e00e2e6)
    _numpy_to_gdal_type = {'uint8': 1, 'int8': 1, 'uint16': 2, 'int16': 3, 'uint32': 4, 'int32': 5, 'float32': 6,
                           'float64': 7, 'complex64': 10, 'compex128': 11}

    _numpy_to_ogr_type = {'uint8': 0, 'int8': 0, 'uint16': 0, 'int16': 0, 'uint32': 0, 'int32': 0, 'float32': 2,
                          'float64': 2}

    # GDAL/OGR underlying attributes
    _gdal_drv = gdal.GetDriverByName('GTiff')
    _ogr_shp_drv = ogr.GetDriverByName('ESRI Shapefile')
    _ogr_geojson_drv = ogr.GetDriverByName('GeoJSON')

    # SetAccess = protected (property decorator)
    geo_grid = protected_property('geo_grid')
    raster_array = protected_property('raster_array')
    crs = protected_property('crs')
    x_origin = protected_property('x_origin')
    y_origin = protected_property('y_origin')
    res = protected_property('res')
    x_size = protected_property('x_size')
    y_size = protected_property('y_size')
    shape = protected_property('shape')
    no_data_value = protected_property('no_data_value')

    def __init__(self, raster, geo_grid: GeoGrid = None, no_data_value=None, crs=None):
        """ RasterMap constructor

        :param raster: raster file or numpy array
        :param geo_grid: GeoGrid instance
        :param no_data_value: data to be regarded as "no_data"
        :param crs: projection used for the raster

        :Example:
        >>>
        """
        check_type(raster, (str, np.ndarray))

        if type(raster) == np.ndarray:
            raster_file = None
            try:
                test_fit = geo_grid.data_fits(raster)
                crs = proj4_from(crs)
                if not test_fit:
                    raise RasterMapError("Input geo grid does not fit raster")
            except AttributeError:
                raise RasterMapError("Geo grid argument has not been set")
            except (ValueError, TypeError):
                raise RasterMapError("Invalid projection: crs='{}'".format(crs))
        else:
            raster_file = raster
            try:
                geo_grid = GeoGrid.from_raster_file(raster)
                crs = proj4_from_raster(raster)
                raster = raster_to_array(raster)
            except RuntimeError:
                raise RasterMapError("Invalid/unknown file '%s'" % raster_file)

        # Set attributes
        self._raster_file = raster_file
        self._geo_grid = geo_grid
        self._raster_array = np.array(raster, dtype='float64')  # Ensure compatibility with NaNs
        self._crs = crs
        self._res = self._geo_grid.res
        self._x_origin = self._geo_grid.geo_transform[0]
        self._y_origin = self._geo_grid.geo_transform[3]
        self._x_size = self._geo_grid.num_x
        self._y_size = self._geo_grid.num_y
        self._shape = self._raster_array.shape
        self._no_data_value = no_data_value

        if no_data_value is not None:
            self._raster_array[self._raster_array == no_data_value] = np.nan  # Use attribute (raster_array) rather than
            # instance (self == ...) to avoid 'recursion' error with decorator above

        # Available filters
        self._filters = {"majority_filter": self._majority_filter, "sieve": self._gdal_sieve}

    @type_assert(filter_name=str)
    def apply_filter(self, filter_name, *args, **kwargs):
        """ Apply filter to raster

        :param filter_name:
        :param args: list of arguments related to filter function
        :param kwargs: list of keyword args related to filter function
        :return:
        """
        return self._filters[filter_name](*args, **kwargs)

    @return_new_instance
    @type_assert(geo_layer=PolygonLayer)
    def clip(self, geo_layer: PolygonLayer, crop=False, all_touched=False):
        """ Clip raster according to GeoLayer polygon(s)

        Keep only points which are inside polygon boundaries
        and crop raster if necessary.
        :param geo_layer: GeoLayer instance
        :param crop: if True, crop raster
        :param all_touched: when True, all cells touched by polygons are considered within
        :return:
        """
        check_proj(geo_layer.crs, self.crs)
        if crop:
            new_raster = self.get_raster_at(geo_layer)
            return new_raster.clip(geo_layer, crop=False, all_touched=all_touched)
        else:
            return self._burn_layer_values(geo_layer, False, all_touched)

    @return_new_instance
    def contour(self, interval, absolute_interval=True, percentile_min=2, percentile_max=98):
        """ Extract contour from raster

        :param interval: interval between contour lines
        :param absolute_interval: relative or absolute interval ?
        :param percentile_min:
        :param percentile_max:
        :return:
        """
        values = self.raster_array[~np.isnan(self.raster_array)]
        v_min, v_max = np.percentile(values, [percentile_min, percentile_max])
        if not absolute_interval:
            interval *= v_max
        contour_range = np.linspace(v_min, v_max, int((v_max - v_min)/interval) + 1)
        new_raster_array = np.zeros(self.raster_array.shape)
        new_raster_array[self.raster_array_without_nans < v_min] = np.mean(values[values < v_min])
        new_raster_array[self.raster_array_without_nans >= v_max] = np.mean(values[values >= v_max])

        for bins in zip(contour_range[0:-1], contour_range[1::]):
            new_raster_array[(self.raster_array_without_nans >= bins[0]) & (self.raster_array_without_nans < bins[
                1])] = np.mean(values[(values >= bins[0]) & (values < bins[1])])

        new_raster_array[self.is_no_data()] = self.no_data_value

        return new_raster_array

    def copy(self):
        return copy.deepcopy(self)

    @return_new_instance
    @type_assert(factor=int, method=str, no_limit=bool)
    def disaggregate(self, factor: int = 1, method: str = 'nearest', no_limit=False):
        """ Disaggregate raster cells

        :param factor: scale factor for disaggregation (number of cells)
        :param method: 'linear' or 'nearest' (default = 'nearest')
        :param no_limit: no limit for disaggregation (default=False)
        :return:
        """
        from scipy.interpolate import RegularGridInterpolator

        upper_limit = np.inf if no_limit else 10**8 / (self.geo_grid.num_x * self.geo_grid.num_y)
        if 1 < factor <= upper_limit:
            new_geo_grid = self.geo_grid.to_res(self.res/factor)
            try:
                interpolator = RegularGridInterpolator((self.geo_grid.lats, self.geo_grid.lons),
                                                       self.raster_array[::-1, :], bounds_error=False,
                                                       method=method)
            except ValueError:
                raise RasterMapError("Method should be 'linear' or 'nearest' but is {}".format(method))

            return interpolator((new_geo_grid.latitude, new_geo_grid.longitude)), new_geo_grid
        else:
            warnings.warn("Invalid factor, factor = 1, or exceeded limit (set no_limit=True). Return copy of object")
            return self.copy()

    @return_new_instance
    @type_assert(geo_layer=PolygonLayer)
    def exclude(self, geo_layer: PolygonLayer, all_touched=False):
        """ Exclude raster cells within GeoLayer polygon(s)

        Keep only points outside from layer boundaries
        :param geo_layer:
        :param all_touched: when True, all cells touched by polygons are regarded as within
        :return:
        """
        check_proj(geo_layer.crs, self.crs)
        return self._burn_layer_values(geo_layer, True, all_touched)

    def gdal_clip(self, extent):
        """ Static method for clipping large raster

        :param extent: array/list as [x_min, y_min, x_max, y_max]
        :return:
        """
        pass

    def gdal_resample(self, factor):
        """ Resampling raster using GDAL

        Unlike the 'disaggregate' method, GDAL uses
        files for resampling (faster for large areas)
        :param factor:

        :return:
        """
        return self._gdal_resample_raster(factor)

    @return_new_instance
    @collection_type_assert(ll_point=dict(collection=(list, tuple), length=2, type=(int, float)), ur_point=dict(
        collection=(list, tuple), type=(int, float), length=2))
    def get_raster_at(self, layer=None, ll_point=None, ur_point=None):
        """ Extract sub-raster in current raster map

        Extract new raster from current raster map
        by giving either a geo lines_ or a new geo-square
        defined by lower-left point (ll_point) and upper
        right point (ur_point) such as for geo grids.
        :param layer: GeoLayer instance
        :param ll_point: tuple of 2 values (lat, lon)
        :param ur_point: tuple of 2 values (lat, lon)
        :return: RasterMap
        """
        if layer is not None:
            check_proj(layer.crs, self.crs)
            ll_point = (layer.bounds[1], layer.bounds[0])  # Warning: (lat, lon) in that order !
            ur_point = (layer.bounds[3], layer.bounds[2])

        try:
            ll_point_r, ll_point_c = self._geo_grid.latlon_to_2d_index(ll_point[0], ll_point[1])
            ur_point_r, ur_point_c = self._geo_grid.latlon_to_2d_index(ur_point[0], ur_point[1])
            return self.raster_array[ur_point_r:ll_point_r + 1, ll_point_c:ur_point_c + 1], \
                self.geo_grid[ur_point_r:ll_point_r + 1, ll_point_c:ur_point_c + 1]
        except GeoGridError:
            raise RasterMapError("Lower left or/and upper right points have not been rightly defined")

    def get_value_at(self, latitude, longitude):
        """ Get single raster value at latitude, longitude

        :param latitude:
        :param longitude:
        :return:
        """
        r, c = self._geo_grid.latlon_to_2d_index(latitude, longitude)
        return self._raster_array[r, c]

    def is_latlong(self):
        return pyproj.Proj(proj4_from(self.crs)).crs.is_geographic

    def is_no_data(self):
        return np.isnan(self.raster_array)

    def max(self):
        """ Return maximum value of raster

        :return:
        """
        return np.nanmax(self.raster_array)

    def mean(self):
        """ Compute mean of raster map

        Mean of raster values
        :return:
        """
        return np.nanmean(self.raster_array)

    def min(self):
        """ Return minimum value of raster

        :return:
        """
        return np.nanmin(self.raster_array)

    def plot(self, ax=None, cmap=None, colorbar=False, colorbar_title=None, **kwargs):
        """ Plot Raster

        :param ax:
        :param cmap:
        :param colorbar:
        :param colorbar_title:
        :param kwargs:
        :return:
        """

        extent = [self.x_origin, self.x_origin + self.res * self.x_size, self.y_origin - self.res * self.y_size,
                  self.y_origin]

        if ax is None:
            _, ax = plt.subplots()

        # Use imshow to plot raster
        img = ax.imshow(self.raster_array, extent=extent, cmap=cmap, vmin=self.min(), vmax=self.max(), **kwargs)

        if colorbar:
            cbar = plt.colorbar(img)
            cbar.ax.set_ylabel(colorbar_title)

        return ax

    @type_assert(field_name=str)
    def polygonize(self, field_name, layer_name="layer", is_8_connected=False):
        """ Convert raster into vector polygon(s)

        :param field_name: name of the corresponding field in the final shape file
        :param layer_name: name of resulting layer
        :param is_8_connected: pixel connectivity used for polygon
        :return:
        """
        check_type(is_8_connected, bool)
        with ShapeTempFile() as out_shp:
            self._gdal_polygonize(out_shp, layer_name, field_name, is_8_connected)
            return PolygonLayer(out_shp, name=layer_name)

    def sum(self):
        """ Compute sum of raster map

        :return:
        """
        return np.nansum(self.raster_array)

    def surface(self):
        """ Return array of raster cell surface values

        :return: numpy ndarray
        """
        surface = compute_surface(self.geo_grid.longitude - self.geo_grid.res / 2, self.geo_grid.longitude +
                                  self.geo_grid.res / 2, self.geo_grid.latitude + self.geo_grid.res / 2,
                                  self.geo_grid.latitude - self.geo_grid.res / 2, self.geo_type,
                                  ellipsoid_from(self.crs))

        return surface

    def to_crs(self, crs):
        """ Reproject raster onto new CRS

        :param crs:
        :return:
        """
        try:
            if not is_equal(self.crs, crs):
                srs = srs_from(crs)
                return self._gdal_warp(srs)
            else:
                return self.copy()
        except ValueError:
            raise RasterMapError("Invalid CRS '%s'" % crs)

    def to_file(self, raster_file, data_type=None):
        """ Save raster to file

        :param raster_file:
        :param data_type: data type
        :return:
        """
        if data_type is None:
            dtype = self._numpy_to_gdal_type[self.data_type]
        else:
            try:
                dtype = self._numpy_to_gdal_type[data_type]
            except KeyError:
                raise RasterMapError("Invalid data type '%s'" % data_type)

        out = array_to_raster(raster_file, self.raster_array_without_nans, self.geo_grid, self.crs, datatype=dtype,
                              no_data_value=self.no_data_value)

        return out

    @classmethod
    def is_equally_referenced(cls, *args):
        """ Check geo referencing equality between rasters

        :param args:
        :return:
        """
        return False not in [raster_1.geo_grid == raster_2.geo_grid for raster_1, raster_2 in zip(args[:-1], args[1::])]

    @property
    def data_type(self):
        return str(self.raster_array.dtype)

    @property
    def geo_type(self):
        if self.is_latlong():
            return 'latlon'
        else:
            return 'equal'

    @property
    def raster_array_without_nans(self):
        array = self.raster_array.copy()
        array[np.isnan(array)] = self.no_data_value
        return array

    @property
    def raster_file(self):
        """ Return underlying raster file

        If raster file does not exist, create a temporary one
        :return:
        """
        if not isfile(self._raster_file):
            with RasterTempFile() as file:
                self._raster_file = file
            self.to_file(self._raster_file)

        return self._raster_file

    @classmethod
    def merge(cls, list_of_raster, bounds=None):
        """ Merge several RasterMap instances

        :param list_of_raster: list of RasterMap class instance(s)
        :param bounds: bounds of the output RasterMap
        :return:
        """
        # TODO: use gdal merge method
        list_of_datasets = []
        for raster_map in list_of_raster:
            list_of_datasets.append(rasterio_open(raster_map.raster_file, 'r'))

        # Merge using rasterio
        array, transform = rasterio_merge(list_of_datasets, bounds=bounds)
        with RasterTempFile() as file:
            with rasterio_open(file, 'w', driver="GTiff", height=array.shape[1], width=array.shape[2], count=1,
                               dtype=array.dtype, crs=list_of_raster[0].crs, transform=transform) as out_dst:
                out_dst.write(array.squeeze(), 1)
            return cls(file, no_data_value=list_of_raster[0].no_data_value)

    def __getitem__(self, key):

        if key.__class__ == type(self):
            key = key.raster_array

        if key.__class__ == slice:
            key = (key, key)

        if key.__class__ == tuple:
            if key[0].__class__ == int and key[1].__class__ == int:
                return self.raster_array.__getitem__(key)
            else:
                try:
                    return self.__class__(self.raster_array.__getitem__(key), self.geo_grid.__getitem__(key),
                                          no_data_value=self.no_data_value, crs=self.crs)
                except IndexError:
                    raise RasterMapError("Invalid indexing")
                except Exception as e:
                    raise RuntimeError("Unknown error while getting data in raster map: {}".format(e))
        elif key.__class__ == np.ndarray:
            return self.raster_array.__getitem__(key)
        else:
            raise RasterMapError("Invalid indexing")

    def __setitem__(self, key, value):

        if type(key) == type(self):
            self.raster_array.__setitem__(key.raster_array, value)
        else:
            try:
                self.raster_array.__setitem__(key, value)
            except IndexError:
                raise RasterMapError("Invalid indexing")
            except Exception as e:
                raise RuntimeError("Unknown error while setting data in raster map: {}".format(e))

        return self

    def __rmul__(self, other):
        return self * other

    def __radd__(self, other):
        return self + other

    def __rsub__(self, other):
        return -self + other

    def __mul__(self, other):
        return self._apply_operator(other, '__mul__', '*')

    def __add__(self, other):
        return self._apply_operator(other, '__add__', '+')

    def __sub__(self, other):
        return self._apply_operator(other, '__sub__', '-')

    def __and__(self, other):
        return self._apply_comparison(other, '__and__', '&')

    def __or__(self, other):
        return self._apply_comparison(other, '__or__', '|')

    def __xor__(self, other):
        return self._apply_comparison(other, '__xor__', '^')

    def __eq__(self, other):
        return self._apply_comparison(other, '__eq__', '==')

    def __ne__(self, other):
        return self._apply_comparison(other, '__ne__', '!=')

    def __gt__(self, other):
        return self._apply_comparison(other, '__gt__', '>')

    def __lt__(self, other):
        return self._apply_comparison(other, '__lt__', '<')

    def __ge__(self, other):
        return self._apply_comparison(other, '__ge__', '>=')

    def __le__(self, other):
        return self._apply_comparison(other, '__le__', '<=')

    @return_new_instance
    def __neg__(self):
        return -self.raster_array

    # @return_new_instance
    def _apply_comparison(self, other, operator_function, operator_str):
        valid_values = np.full(self.raster_array.shape, False)
        if isinstance(other, RasterMap):
            other = other.raster_array
        try:
            valid_values[(~np.isnan(self.raster_array)) & (~np.isnan(other))] = True
            valid_values[valid_values] = self.raster_array[valid_values].__getattribute__(operator_function)(
                other[valid_values])
            return valid_values
        except (TypeError, IndexError):
            valid_values[~np.isnan(self.raster_array)] = True
            valid_values[valid_values] = self.raster_array[valid_values].__getattribute__(operator_function)(other)
            return valid_values
        except Exception as e:
            raise RasterMapError("Comparison for '{}' has failed ({})".format(operator_str, e))

    @return_new_instance
    def _apply_operator(self, other, operator_function, operator_str):
        if isinstance(other, RasterMap):
            if self.geo_grid == other.geo_grid:
                return self.raster_array.__getattribute__(operator_function)(other.raster_array)
            else:
                raise RasterMapError("Raster maps are not defined on the same geo grid")
        else:
            try:
                return self.raster_array.__getattribute__(operator_function)(other)
            except TypeError:
                raise RasterMapError("Unsupported operand type(s) for {}: '{}' and '{}'".format(operator_str,
                                                                                                type(self).__name__,
                                                                                                type(other).__name__))
            except ValueError:
                raise RasterMapError("No match for operand {} between '{}' and '{}'".format(operator_str,
                                                                                            type(self).__name__,
                                                                                            type(other).__name__))
            except Exception as e:
                raise RuntimeError("Unexpected error when applying {} between '{}' and '{}': {}"
                                   .format(operator_str, type(self).__name__, type(other).__name__, e))

    def _burn_layer_values(self, geo_layer, mask, all_touched):
        """ Burn raster values inside or outside layer

        :param geo_layer:
        :param mask:
        :param all_touched:
        :return:
        """
        layer = geo_layer.copy()
        layer["burn_value"] = 1
        layer = layer.to_array(self.geo_grid, "burn_value", all_touched=all_touched)
        new_raster_array = self.raster_array.copy()
        if mask:
            new_raster_array[layer == 1] = np.nan
        else:
            new_raster_array[layer != 1] = np.nan

        return new_raster_array

    def _gdal_polygonize(self, out_shp, layer_name, field_name, is_8_connected):
        """ Polygonize raster using GDAL

        :param out_shp: shapefile
        :param layer_name:
        :param field_name:
        :param is_8_connected:
        :return:
        """
        connectivity = "8CONNECTED=%d" % (8 if is_8_connected else 4)

        with GdalOpen(self.raster_file) as src_ds:
            src_band = src_ds.GetRasterBand(1)
            dst_ds = self._ogr_shp_drv.CreateDataSource(out_shp)
            dst_layer = dst_ds.CreateLayer(layer_name, srs_from(self.crs))

            fd = ogr.FieldDefn(field_name, self._numpy_to_ogr_type[self.data_type])
            dst_layer.CreateField(fd)

            gdal.Polygonize(src_band, src_band.GetMaskBand(), dst_layer, 0, [connectivity])

    @gdal_decorator()
    def _gdal_resample_raster(self, out_raster, factor):
        """ Resample raster using GDAL utility

        :param factor:
        :return:
        """
        with GdalOpen(self.raster_file) as source_ds:
            dst_ds = self._gdal_drv.Create(out_raster, self.x_size * factor, self.y_size * factor, 1,
                                           source_ds.GetRasterBand(1).DataType)
            resample_geo_transform = (self.x_origin, self.res / factor, 0, self.y_origin, 0, -self.res / factor)
            dst_ds.SetGeoTransform(resample_geo_transform)
            dst_ds.SetProjection(wkt_from(self.crs))
            gdal.RegenerateOverview(source_ds.GetRasterBand(1), dst_ds.GetRasterBand(1), 'mode')

    @gdal_decorator()
    def _gdal_warp(self, output_raster, srs):
        with GdalOpen(self.raster_file) as src_ds:
            gdal.Warp(output_raster, src_ds, dstSRS=srs)

    @gdal_decorator()
    def _gdal_sieve(self, out_raster, *args, **kwargs):
        """ Sieve filter using gdal

        :param args:
        :param kwargs:
        :return:
        """
        # Destination dataset
        dst_ds = self._clone_gdal_dataset(out_raster)

        # Apply sieve filter
        with GdalOpen(self.raster_file) as source_ds:
            gdal.SieveFilter(source_ds.GetRasterBand(1), None, dst_ds.GetRasterBand(1), *args, **kwargs)

    def _majority_filter(self):
        pass

    def _clone_gdal_dataset(self, out_raster):
        """ Create GDAL dataset based on raster map properties

        :param out_raster: out raster file
        :return:
        """
        with GdalOpen(self.raster_file) as source_ds:
            dst_ds = self._gdal_drv.Create(out_raster, self.x_size, self.y_size, 1, source_ds.GetRasterBand(1).DataType)
            dst_ds.SetGeoTransform(self.geo_grid.geo_transform)
            dst_ds.SetProjection(wkt_from(self.crs))

            if self.no_data_value is not None:
                dst_band = dst_ds.GetRasterBand(1)
                dst_band.SetNoDataValue(self.no_data_value)

        return dst_ds


class ClassificationMap(RasterMap):
    """ Classification raster map

    Store classification values as integers
    """

    def __init__(self, raster, geo_grid, crs):
        super().__init__(raster, geo_grid, crs=crs, no_data_value=None)

        # Update available filters
        self._filters.update({"sieve": self._gdal_sieve})

    @staticmethod
    @type_assert(raster_map=RasterMap)
    def from_raster_map(raster_map):
        """ Build classification raster from RasterMap instance

        :param raster_map:
        :return:
        """
        pass


class DigitalElevationModel(RasterMap):
    """ Digital Elevation SubModel class

    Store Digital Elevation SubModel
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def compute_slope(self, slope_format="percent"):
        """ Compute slope from DEM

        :return:
        """
        slope_format = check_string(slope_format, {'percent', 'degree'})
        return self._compute_slope(slope_format)

    def compute_aspect(self):
        """ Compute aspect

        :return:
        """
        return self._compute_aspect()

    # TODO: relate following method to functions in "topography" module
    def get_horizon(self):
        pass

    @gdal_decorator(no_data_value=-9999)
    def _compute_slope(self, out_raster, slope_format):

        with GdalOpen(self.raster_file) as src_ds:
            options = gdal.DEMProcessingOptions(slopeFormat=slope_format)
            gdal.DEMProcessing(out_raster, src_ds, 'slope', options=options)

    @gdal_decorator(no_data_value=-9999)
    def _compute_aspect(self, out_raster):

        with GdalOpen(self.raster_file) as src_ds:
            gdal.DEMProcessing(out_raster, src_ds, 'aspect')

    @staticmethod
    @type_assert(bounds=tuple, product=str, margin=(int, float))
    def from_online_srtm_database(bounds, path_to_dem_file=DEFAULT_OUTPUT, product="SRTM1", margin=0,
                                  no_data_value=-32768):
        """ Import DEM tile from SRTM3 or SRTM1 online dataset

        Based on "elevation" module. Be careful that at the moment, SRTM3 product
        does not seem to work properly.
        :param bounds:
        :param path_to_dem_file:
        :param product: "SRTM1" or "SRTM3"
        :param margin: margin (in %) around DEM
        :param no_data_value: no data filling value (default = -32768)
        :return:
        """
        from subprocess import CalledProcessError
        import elevation

        try:
            check_string(product, {'SRTM1', 'SRTM3'})
        except ValueError as e:
            raise DigitalElevationModelError("Invalid product name '%s': %s" % (product, e))

        try:
            elevation.clip(bounds, output=path_to_dem_file, margin="%s" % (margin/100), product=product)
        except CalledProcessError as e:
            raise DigitalElevationModelError("Internal subprocess error: %s" % e)
        except (ValueError, TypeError, KeyError) as e:
            raise DigitalElevationModelError("Invalid input argument: %s" % e)

        # Return instance of DigitalElevationModel
        return DigitalElevationModel(path_to_dem_file, no_data_value=no_data_value)

    @staticmethod
    def from_cgiar_online_database(bounds, margin=0, max_tiles=4, no_data_value=-32768):
        """ Import DEM tile from CGIAR-CSI SRTM3 database (V4.1)

        :param bounds: bounds of the image --> (x_min, y_min, x_max, y_max)
        :param margin: margin (in %) around DEM
        :param max_tiles: max number of tiles to download
        :param no_data_value: no data filling value (default = -32768)
        :return:
        """
        srtm_lon = np.arange(-180, 185, 5)
        srtm_lat = np.arange(60, -65, -5)
        bounds = (bounds[0] - (bounds[2] - bounds[0]) * margin/100, bounds[1] - (bounds[3] - bounds[1]) * margin/100,
                  bounds[2] + (bounds[2] - bounds[0]) * margin/100, bounds[3] + (bounds[3] - bounds[1]) * margin/100)
        x_min, x_max = np.digitize(bounds[0], srtm_lon, right=True), np.digitize(bounds[2], srtm_lon, right=True)
        y_min, y_max = np.digitize(bounds[3], srtm_lat), np.digitize(bounds[1], srtm_lat)

        if (x_max - x_min) * (y_max-y_min) > max_tiles:
            raise DigitalElevationModelError("Too much tiles to download (>%d)" % max_tiles)

        list_of_tiles = []

        for x in range(int(x_min), int(x_max) + 1):
            for y in range(int(y_min), int(y_max) + 1):
                tile_temp_file = _download_srtm_tile("srtm_%02d_%02d" % (x, y))
                list_of_tiles.append(DigitalElevationModel(tile_temp_file, no_data_value=no_data_value))

        # Merge DEMs
        return DigitalElevationModel.merge(list_of_tiles, bounds)


def _download_srtm_tile(tile_name):
    """ Download and extract SRTM tile archive

    :param tile_name:
    :return:
    """
    zip_name, tif_name = tile_name + ".zip", tile_name + '.tif'
    url = os.path.join(CGIAR_URL, zip_name)
    temp_srtm_zip = os.path.join(tempfile.gettempdir(), zip_name)
    temp_srtm_dir = os.path.join(tempfile.gettempdir(), tile_name)

    # Download tile
    try:
        urlretrieve(url, temp_srtm_zip)
    except URLError as e:
        raise RuntimeError("Unable to fetch data at '%s': %s" % (url, e))

    # Extract GeoTiff
    archive = ZipFile(temp_srtm_zip, 'r')
    archive.extractall(temp_srtm_dir)
    archive.close()

    return os.path.join(temp_srtm_dir, tif_name)


if __name__ == "__main__":
    test = RasterMap("/home/benjamin/Documents/Data/Resource rasters/Solar maps/Monthly GHI sum/monthly_GHI_01.tif")
    m0 = test[10:200]
    m = test[10:200, 10:200]
    m2 = test[test > 1278]
    m3 = test[7, 8]
    print((m0 == m).all())
    print("done")



# -*- coding: utf-8 -*-

""" Manage database connections and instances (postgis, etc.)

This module intends to make data reading from spatial databases with gistools easy
"""
import warnings

from geopandas import GeoDataFrame
from gistools.exceptions import SpatialDatabaseWarning, SpatialDatabaseError, GeoLayerError
from gistools.layer import PolygonLayer, PointLayer, LineLayer
from sqlalchemy import create_engine, inspect, MetaData
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, Session
from utils.check import lazyproperty, check_string

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2019, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


class SpatialDatabase:
    """ Spatial database class based on Postgresql

    """

    session = None
    layer_class = {'point': PointLayer, 'line': LineLayer, 'polygon': PolygonLayer}

    def __init__(self, user, passwd, ip, db_name):
        """ Initialize spatial database from db string connection

        :param user: database user
        :param passwd: database password
        :param ip: database server ip address
        :param db_name: database name
        """

        self.engine = create_engine('postgresql://%s:%s@%s/%s' % (user, passwd, ip, db_name), echo=True)
        self.db_name = db_name

    def automap(self):
        """ Automatically map database into sqlalchemy ORM model

        """

        metadata = MetaData()
        metadata.reflect(self.engine, only=self.table_names)
        base = automap_base(metadata=metadata)
        base.prepare()

        for table_name in self.table_names:
            try:
                self.__setattr__(table_name.capitalize(), base.classes.__getattr__(table_name))
            except AttributeError:
                warnings.warn("Invalid or missing table '%s'" % table_name, SpatialDatabaseWarning)

        self.session = Session(self.engine)

    def table_to_layer(self, table_name, geom_type=None, bounds=None, polygon_extent=None):
        """ Convert table from database to GeoLayer instance

        :param table_name: name of table
        :param geom_type: geometry type
        :param bounds: bounding box (x_min, y_min, x_max, y_max)
        :param polygon_extent: shapely polygon
        :return:
        """
        if bounds is not None and polygon_extent is None:
            sql_string = f'SELECT * FROM {table_name} WHERE {table_name}.geom && ST_MakeEnvelope({bounds[0]}, ' \
                         f'{bounds[1]}, {bounds[2]}, {bounds[3]})'
        elif polygon_extent is not None and bounds is None:
            sql_string = f'SELECT * FROM {table_name} WHERE ST_Within({table_name}.geom, {polygon_extent})'
        else:
            sql_string = f'SELECT * FROM {table_name}'

        df = GeoDataFrame.from_postgis(sql_string, self.engine)

        if table_name in self.table_names and geom_type is None:
            try:
                layer = PolygonLayer(df, name=table_name)
            except GeoLayerError:
                try:
                    layer = LineLayer(df, name=table_name)
                except GeoLayerError:
                    layer = PointLayer(df, name=table_name)
        elif table_name in self.table_names and geom_type is not None:
            try:
                geom_type = check_string(geom_type, ("point", "line", "polygon"))
            except ValueError:
                raise SpatialDatabaseError("Invalid geometry type '%s'. Must be 'point', 'line' or 'polygon'." %
                                           geom_type)
            layer = self.layer_class[geom_type](df, name=table_name)
        else:
            raise SpatialDatabaseError("No table named '%s' in database '%s'" % (table_name, self.db_name))

        return layer

    @property
    def table_names(self):
        inspector = inspect(self.engine)
        return inspector.get_table_names()



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

# __all__ = []
# __version__ = '0.1'
from abc import abstractmethod

import networkx as nx
import numpy as np
from geopandas import GeoDataFrame
from numba import jit, float64
from shapely.geometry import Point, LineString
from shapely.ops import linemerge
from utils.toolset import split_list_by_index
from utils.check.descriptor import protected_property
from utils.check.type import check_type, type_assert, check_type_in_collection
from utils.check.value import check_string

from gistools.coordinates import r_tree_idx
from gistools.exceptions import EdgeError, NetworkError, RoadError, RoadNodeError
from gistools.geometry import connect_lines_to_point, centroid, intersects
from gistools.layer import return_new_instance, LineLayer, PointLayer

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2018, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


# Division by zero
np.seterr(divide='ignore')

# Constants
SPEED_RATIO = {'m/s': 1, 'km/h': 3.6}
TIME_FORMAT = {'s': 1, 'm': 1/60, 'h': 1/3600}


def multi_edges(graph):
    """ Return multiple edges between 2 nodes

    :return: list of edge IDs that connect the same nodes
    """
    multiple_edges = []
    for node in graph.nodes():
        for neighbor in graph.neighbors(node):
            if graph.number_of_edges(node, neighbor) > 1:
                edge_id = [i for i in graph[node][neighbor]]
                if edge_id not in multiple_edges:
                    multiple_edges.append(edge_id)

    return multiple_edges


def remote_edges(graph):
    """ Return remote edges

    Remote edges are not connected to anything
    :return: list of edge IDs that are remote
    """
    rm_edges = []
    for u, v in graph.edges():
        if len(list(graph.neighbors(u))) == 1 and len(list(graph.neighbors(v))) == 1:
            rm_edges.extend(list(graph[u][v]))

    return rm_edges


def remote_nodes(graph):
    """ Return remote nodes

    Remote nodes are isolated nodes that are not connected to anything
    :param graph:
    :return:
    """
    return [u for u in graph.nodes() if len(list(graph.neighbors(u))) == 0]


def self_loops(graph):
    """ Return self-loop edges

    :return: list of edge IDs that are self-loops
    """
    slf_loops = []
    for self_loop in nx.selfloop_edges(graph, keys=True):
        slf_loops.append(self_loop[2])

    return slf_loops


def find_all_disconnected_edges_and_fix(edges, tolerance, method):
    """ Find all disconnected edges of network and fix it

    :param edges: Edge class instance
    :param tolerance: tolerance for considering "disconnected" an edge
    :param method: fix method
    :return:
    """
    while "There is still disconnected edges":
        new_edges = edges.find_disconnected_islands_and_fix(tolerance=tolerance, method=method)
        if len(new_edges) < len(edges):
            edges = new_edges
        else:
            break

    return new_edges


class Edge(LineLayer):
    """ Edge class

    Class for implementing edges
    in a geo network
    """
    from_to = protected_property('from_to')

    DEFAULT_DIRECTION = "two-ways"

    def __init__(self, edges, *args, **kwargs):
        """ Edge class constructor

        :param edges: geo-like file (e.g. shapefile) or geopandas data frame
        """
        super().__init__(edges, *args, **kwargs)

        # Set Edge specific attributes
        if "direction" not in self.attributes():
            self["direction"] = self.DEFAULT_DIRECTION
            # self._gpd_df["direction"] = [self.DEFAULT_DIRECTION] * len(self)  # Set default direction (not directed)

        # Simplified edges
        self._from_to = [((geom.coords.xy[0][0], geom.coords.xy[1][0]), (geom.coords.xy[0][-1], geom.coords.xy[1][
            -1])) for geom in self.geometry]

        # Corresponding undirected multi-graph
        self._graph = nx.MultiGraph()
        self._graph.add_edges_from([(from_to[0], from_to[1], _id) for _id, from_to in enumerate(self.from_to)])

        # Override point layer class attribute (To Edge is associated Node)
        self._point_layer_class = Node

    def _check_attr_and_dic(self, attr_name, dic):

        if attr_name not in self.attributes():
            raise EdgeError("Unknown attribute name '%s'" % attr_name)

        if any([val not in list(set(self[attr_name])) for val in dic.keys()]):
            raise EdgeError("Invalid key in '%s'" % dic)

    def find_disconnected_islands_and_fix(self, tolerance=None, method="delete"):
        """ Find disconnected components in network

        Find disconnected components/islands graphs in multi-graph
        and apply method (fix/reconnect, keep, delete) with respect
        to a given tolerance
        :param tolerance:
        :param method:
        :return:
        """
        method = check_string(method, {'reconnect_and_delete', 'reconnect_and_keep', 'delete'})
        sub_graphs = list(nx.connected_component_subgraphs(self._graph))
        main_component = max(sub_graphs, key=len)
        sub_graphs.remove(main_component)

        idx_edge = []
        if method == "delete":
            for graph in sub_graphs:
                for edge in graph.edges:
                    try:
                        idx_edge.append(self.from_to.index((edge[0], edge[1])))
                    except ValueError:
                        idx_edge.append(self.from_to.index((edge[1], edge[0])))

        elif method == 'reconnect_and_delete':
            pass
        elif method == 'reconnect_and_keep':
            pass

        return self.drop(self.index[idx_edge])

        # TODO: implement island reconnection and tolerance (see Edge.reconnect() method)

    @return_new_instance
    def get_path(self, path):

        edge_path = GeoDataFrame(columns=self.attributes(), crs=self.crs)

        for i in range(len(path) - 1):
            try:
                edge_path = edge_path.append(self._gpd_df.loc[self._from_to.index((path[i], path[i + 1]))],
                                             ignore_index=True)
            except ValueError:
                edge_path = edge_path.append(self._gpd_df.loc[self._from_to.index((path[i + 1], path[i]))],
                                             ignore_index=True)

        return edge_path

    def get_end_nodes(self):
        """ Get end nodes of edges

        Get end nodes, that is edge's end that does not connect
        to another edge
        :return: Node layer and list of edge IDs
        """
        end = [(node, list(self._graph.edges(node, keys=True))[0]) for node in self._graph.nodes() if len(
            self._graph.edges(node)) == 1]
        return self._point_layer_class.from_gpd(geometry=[Point(n[0]) for n in end], crs=self.crs),\
            [n[1][2] for n in end]

    def get_nodes(self):
        """ Get nodes from edges

        :return:
        """
        return self._point_layer_class.from_gpd(geometry=[Point(node) for node in self._graph.nodes()], crs=self.crs)

    def get_multi_edges(self):
        """ Get multiple edges between 2 nodes

        :return: list of edge IDs that connect the same nodes
        """
        return multi_edges(self._graph)

    def get_remote_edges(self):
        """ Get remote edges

        Remote edges are not connected to anything
        :return: list of edge IDs that are remote
        """
        return remote_edges(self._graph)

    def get_self_loops(self):
        """ Get self-loop edges

        :return: list of edge IDs that are self-loops
        """
        return self_loops(self._graph)

    @return_new_instance
    def get_simplified(self):
        """ Return simplified edge

        Return simplified Edge instance, that is only with starting
        and ending coordinates of each road segment
        :return:
        """
        return GeoDataFrame(self._gpd_df.copy(), geometry=[LineString(from_to) for from_to in self._from_to],
                            crs=self.crs)

    def get_single_edges(self):
        """ Get single edges

        Retrieve single edges, that is from intersection to intersection.
        Typically, when a node has only 2 neighbors, the corresponding
        edges can be merged into a new one.
        :return: list of list of edge IDs representing unique elements
        """
        to_merge = []
        all_connected_edges = []

        for u, v, edge_id in self._graph.edges(keys=True):
            if edge_id not in all_connected_edges:
                connected_nodes = [u, v]
                connected_edges = [edge_id]
                for node in [u, v]:
                    previous_node = node
                    while "There are neighbors to connect":
                        next_node = [n for n in self._graph.neighbors(previous_node) if n not in connected_nodes]
                        if not next_node or len(next_node) > 1:
                            break
                        else:
                            connected_edges.extend(list(self._graph[previous_node][next_node[0]]))
                            connected_nodes.extend(next_node)
                            previous_node = next_node[0]  # previous node is now the next node in the chain

                all_connected_edges.extend(connected_edges)
                to_merge.append(connected_edges)

        return to_merge

    @return_new_instance
    def merge2(self):
        """ Merge edges from intersection to intersection

        Merge with respect to single entities, that is from
        intersecting node to intersecting node (node with
        more than 2 neighbors).
        :return:
        """
        single_edges = self.get_single_edges()
        geometry, rows = [], []

        for line in single_edges:
            geometry.append(linemerge(self.geometry[line].values))
            rows.append(self._gpd_df.iloc[line[-1]])

        return GeoDataFrame(rows, geometry=geometry, crs=self.crs)

    @return_new_instance
    def reconnect(self, tolerance):
        """ Reconnect disconnected edges with respect to tolerance

        :param tolerance: min distance (in m) for reconnection
        :return:
        """
        # TODO: link with Edge.find_disconnected_islands_and_fix() method
        outdf = self._gpd_df.copy()
        nodes, edge_idx = self.get_end_nodes()
        nearest_nodes, node_idx = nodes.nearest_neighbors(nodes, tolerance)
        connected_nodes = []

        for n, n_nodes in enumerate(nearest_nodes):
            n_idx = [node for node in node_idx[n] if node not in connected_nodes]
            if len(n_idx) > 1:
                e_idx = [edge_idx[i] for i in n_idx]
                connected_nodes.extend([i for i in n_idx if i not in connected_nodes])
                # Use outdf.geometry to ensure that changes are saved
                new_geometry = connect_lines_to_point(outdf.geometry[e_idx], centroid(n_nodes.geometry))
                for edge, geom in zip(e_idx, new_geometry):
                    outdf.loc[edge, "geometry"] = geom

        return outdf

    @type_assert(attribute_name=str, direction_dic=dict)
    def set_direction(self, attribute_name, direction_dic):
        """ Set edge direction

        :param attribute_name: layer attribute from which direction must be derived
        :param direction_dic: (valid direction values: "two-ways", "one-way", "reverse", None)
        :return:
        """
        self._check_attr_and_dic(attribute_name, direction_dic)

        for key in direction_dic.keys():
            if direction_dic[key] not in ['two-ways', 'one-way', 'reverse', None]:
                raise EdgeError("'%s' is not a valid direction value" % direction_dic[key])
            self._gpd_df.loc[self[attribute_name] == key, "direction"] = direction_dic[key]

    def split_at_ending_edges(self):
        """ Split edge on which ends another edge

        :return:
        """
        nodes, edge_idx = self.get_end_nodes()
        splitting_nodes = nodes[[True if intersects(geom, self.geometry, self.r_tree_idx).count(True) > 1 else False for
                                 geom in nodes.geometry]]

        return self.split_at_points(splitting_nodes)

    def split_at_underlying_points(self, location, *args):
        """ Override parent class method

        Split corresponding attributes in addition
        to layer, i.e.
        :param location:
        :return:
        """
        output = super().split_at_underlying_points(location)
        if len(args) == 0:
            return output

        outputs = [output]
        for attr in args:
            split_attr = []
            for n, a in enumerate(attr):
                break_idx = [loc[1] for loc in location if loc[0] == n]
                if len(break_idx) == 0:
                    split_attr.append(a)
                else:
                    split_attr.extend(split_list_by_index(a, break_idx, include=False))
            outputs.append(split_attr)

        return tuple(outputs)

    @property
    def direction(self):
        return self["direction"]

    @property
    def from_node(self):
        return [coords[0] for coords in self._from_to]

    @property
    def to_node(self):
        return [coords[1] for coords in self._from_to]


class Node(PointLayer):
    """ Node class

    Class for implementing a set of nodes
    in a geo network
    """

    def __init__(self, nodes, *args, **kwargs):
        """ Node class constructor

        :param nodes:
        """
        super().__init__(nodes, *args, **kwargs)

    @type_assert(edges=Edge)
    def which_edge(self, edges):
        """ Find to which edge belongs nodes

        :param edges:
        :return:
        """
        return [i for node in self.geometry for i, from_to in enumerate(edges.from_to) if (node.x, node.y) in from_to]


class Road(Edge):
    """ Road class

    Class for implementing a set of roads in a geo network
    """

    DEFAULT_MAX_SPEED = 25
    DEFAULT_ROLLING_COEFFICIENT = 0.01
    DEFAULT_ROLLOVER_CRITERION = 0.15

    def __init__(self, roads, *args, **kwargs):
        """ Road class constructor

        :param roads:
        :param args:
        :param kwargs:
        """
        super().__init__(roads, *args, **kwargs)

        # Max speed on road segment
        if "max_speed" not in self.attributes():
            self._gpd_df["max_speed"] = [self.DEFAULT_MAX_SPEED] * len(self)

        # Rolling coefficient on road segment
        if "rolling_coefficient" not in self.attributes():
            self._gpd_df["rolling_coefficient"] = [self.DEFAULT_ROLLING_COEFFICIENT] * len(self)

        # Rollover criterion on road segment
        if "rollover_criterion" not in self.attributes():
            self._gpd_df["rollover_criterion"] = [self.DEFAULT_ROLLOVER_CRITERION] * len(self)

        # Override point layer class
        self._point_layer_class = RoadNode

    @type_assert(attr_name=str, speed_dic=dict)
    def set_max_speed(self, attr_name, speed_dic, speed_format='km/h'):
        """ Set road max allowed speed

        :param attr_name:
        :param speed_dic:
        :param speed_format:
        :return:
        """
        check_string(speed_format, ('m/s', 'km/h'))
        self._check_attr_and_dic(attr_name, speed_dic)

        for key in speed_dic.keys():
            if not isinstance(speed_dic[key], (float, int)):
                raise RoadError("Speed value must be numeric but is '%s'" % type(speed_dic[key]))
            self._gpd_df.loc[self[attr_name] == key, "max_speed"] = speed_dic[key] / SPEED_RATIO[speed_format]

    @type_assert(attr_name=str, rolling_coeff_dic=dict)
    def set_rolling_coefficient(self, attr_name, rolling_coeff_dic):
        """ Set road rolling coefficient

        :param attr_name:
        :param rolling_coeff_dic:
        :return:
        """
        self._check_attr_and_dic(attr_name, rolling_coeff_dic)

        for key in rolling_coeff_dic.keys():
            if not isinstance(rolling_coeff_dic[key], float):
                raise RoadError("Rolling coefficient must be a float but is '%s'" % type(rolling_coeff_dic[key]))
            self._gpd_df.loc[self[attr_name] == key, "rolling_coefficient"] = rolling_coeff_dic[key]

    @type_assert(attr_name=str, rollover_criterion_dic=dict)
    def set_rollover_criterion(self, attr_name, rollover_criterion_dic):
        """ Set road rollover criterion

        :param attr_name:
        :param rollover_criterion_dic:
        :return:
        """
        self._check_attr_and_dic(attr_name, rollover_criterion_dic)

        for key in rollover_criterion_dic.keys():
            if not isinstance(rollover_criterion_dic[key], float):
                raise RoadError("Rollover criterion must be a float but is '%s'" % type(rollover_criterion_dic[key]))
            self._gpd_df.loc[self[attr_name] == key, "rollover_criterion"] = rollover_criterion_dic[key]

    @property
    def max_speed(self):
        return self["max_speed"]

    @property
    def rolling_coefficient(self):
        return self["rolling_coefficient"]

    @property
    def rollover_criterion(self):
        return self["rollover_criterion"]


class RoadNode(Node):
    """ RoadIntersection class

    Class for implementing road intersections (inherits from Node)
    """

    _check_attr_and_dic = Edge._check_attr_and_dic

    DEFAULT_MAX_SPEED = 0

    def __init__(self, nodes, *args, **kwargs):
        super().__init__(nodes, *args, **kwargs)

        if "max_speed" not in self.attributes():
            self._gpd_df["max_speed"] = [self.DEFAULT_MAX_SPEED] * len(self)

    def set_max_speed(self, attr_name, speed_dic, speed_format='km/h'):
        """ Set max allowed speed at intersection

        :param attr_name:
        :param speed_dic:
        :param speed_format:
        :return:
        """
        check_string(speed_format, ('m/s', 'km/h'))
        self._check_attr_and_dic(attr_name, speed_dic)

        for key in speed_dic.keys():
            if not isinstance(speed_dic[key], (float, int)):
                raise RoadNodeError("Speed value must be numeric but is '%s'" % type(speed_dic[key]))
            self._gpd_df.loc[self[attr_name] == key, "max_speed"] = speed_dic[key] / SPEED_RATIO[speed_format]

    @property
    def max_speed(self):
        return self["max_speed"]


class Network:
    """ Network base class

    Use this class to implement sub-class network
    geometry (e.g. from shapefile) and apply
    corresponding tools
    """
    edges = protected_property("edges")
    nodes = protected_property("nodes")
    _graph = None

    def __init__(self, edges, nodes, match_edge_nodes=True, tolerance=1):
        """ Network class constructor

        :param edges: Edge instance
        :param nodes: Node instance
        :param match_edge_nodes: Boolean --> match edge nodes with respect to tolerance
        :param tolerance: distance tolerance for considering nodes and edge nodes the same (in m)
        """
        check_type(edges, Edge, nodes, Node)
        self._edges = edges
        self._nodes = nodes

        # Retrieve edge nodes corresponding to nodes
        if match_edge_nodes:
            edge_nodes = self._edges.get_nodes()
            distance, nn = nodes.distance_and_nearest_neighbor(edge_nodes)
            self._nodes["geometry"] = [edge_nodes.geometry[n] for n in nn]
            self._nodes = self._nodes[distance <= tolerance]

    def _get_nearest_edge_node(self):
        """ Get nearest node from edge

        :return:
        """
        nodes = self._edges.get_nodes()
        idx = r_tree_idx(nodes.geometry)
        edge_nodes = []
        for geom in self.nodes.geometry:
            nn = list(idx.nearest(geom.bounds, 1))
            edge_nodes.append(nodes.geometry[nn[0]])

        return Node.from_gpd(geometry=edge_nodes, crs=self._edges.crs)

    @abstractmethod
    def build_graph(self, *args, **kwargs):
        pass

    def get_self_loops(self):
        """ Get self-loop edges in network

        :return: list of edge IDs
        """
        return self_loops(self.graph)

    def get_remote_edges(self):
        """ Get remote edges in network

        :return: list of edge IDs
        """
        return remote_edges(self.graph)

    def get_multi_edges(self):
        """ Get multi-edges in network

        :return: list of list of edge IDs
        """
        return multi_edges(self.graph)

    def get_minimum_distance_to_network(self, layer):
        """ get minimum distance from given layer to network

        :param layer:
        :return:
        """
        distance_to_edge = layer.distance(self.edges)
        distance_to_node = layer.distance(self.nodes)

        return np.minimum(distance_to_edge, distance_to_node)

    @type_assert(node_start=Point, node_end=Point)
    def get_shortest_path(self, node_start, node_end):
        """ Get shortest path between 2 nodes using Dijkstra algorithm

        :param node_start: shapely Point
        :param node_end: shapely Point
        :return: Edge instance of the path
        """
        if node_start not in self.nodes.geometry or node_end not in self.nodes.geometry:
            raise EdgeError("Either source or destination node is invalid")

        node_start = (node_start.x, node_start.y)
        node_end = (node_end.x, node_end.y)

        if node_start == node_end:
            return []  # Empty path

        try:
            path = nx.dijkstra_path(self.graph, node_start, node_end)
        except nx.NetworkXNoPath:
            print("No available path between node {} and node {}".format(node_start, node_end))
            return None
        else:
            return self._edges.get_path(path)

    @type_assert(node_start=Point, node_end=Point)
    def get_shortest_path_length(self, node_start, node_end, method: str = "networkx"):
        """ Get dijkstra shortest path length

        :param node_start: shapely Point
        :param node_end: shapely Point
        :param method: {'internal', 'networkx'}
        :return: length of path in m
        """
        check_string(method, ("internal", "networkx"))
        if method == "internal":
            edge_path = self.get_shortest_path(node_start, node_end)
            length = 0
            if edge_path is not None and edge_path != []:
                for edge in edge_path.geometry:
                    length += edge.length
            elif edge_path is None:
                return None
        else:
            node_start = (node_start.x, node_start.y)
            node_end = (node_end.x, node_end.y)
            length = nx.dijkstra_path_length(self.graph, node_start, node_end)

        return length

    def get_all_shortest_paths(self):
        """ Get shortest paths between all graph nodes

        :return:
        """
        return nx.all_pairs_dijkstra_path(self.graph)

    def get_all_shortest_path_lengths(self):
        """ Get shortest path lengths between all graph nodes

        :return:
        """
        return nx.all_pairs_dijkstra_path_length(self.graph)

    @type_assert(source_node=Point)
    def get_all_shortest_paths_from_source(self, source_node):
        """ Get all paths from one source node using Dijkstra

        :param source_node:
        :return:
        """
        if source_node not in self.nodes.geometry:
            raise NetworkError("Source node is invalid")

        source_node = (source_node.x, source_node.y)

        return nx.single_source_dijkstra_path(self.graph, source_node)

    @type_assert(source_node=Point)
    def get_all_shortest_path_lengths_from_source(self, source_node):
        """

        :param source_node:
        :return:
        """
        if source_node not in self.nodes.geometry:
            raise NetworkError("Source node is invalid")

        source_node = (source_node.x, source_node.y)

        return nx.single_source_dijkstra_path_length(self.graph, source_node)

    @type_assert(source_node=Point)
    def get_shortest_paths_from_source(self, source_node, target_nodes):
        """ Get multiple shortest paths from single source using Dijkstra

        :param source_node:
        :param target_nodes:
        :return:
        """
        try:
            check_type_in_collection(target_nodes, Point)
        except TypeError:
            raise NetworkError("'%s' must be a collection of Point instances" % target_nodes)

        paths = []
        all_paths = self.get_all_shortest_paths_from_source(source_node)
        for target in target_nodes:
            target = (target.x, target.y)
            if target in all_paths.keys():
                paths.append(self._edges.get_path(all_paths[target]))

        return paths

    @type_assert(source_node=Point)
    def get_shortest_path_lengths_from_source(self, source_node, target_nodes):
        """

        :param source_node:
        :param target_nodes:
        :return:
        """
        try:
            check_type_in_collection(target_nodes, Point)
        except TypeError:
            raise NetworkError("'%s' must be a collection of Point instances" % target_nodes)

        path_lengths = []
        all_path_lengths = self.get_all_shortest_path_lengths_from_source(source_node)
        for target in target_nodes:
            target = (target.x, target.y)
            if target in all_path_lengths.keys():
                path_lengths.append(all_path_lengths[target])

        return path_lengths

    def get_shortest_path_matrix(self):
        """ Get shortest path matrix

        Compute shortest path length between all
        starting and ending nodes
        :return:
        """
        shortest_path = np.full((len(self.nodes), len(self.nodes)), np.nan)
        edge_nodes = self._get_nearest_edge_node()
        for i, geom_from in enumerate(edge_nodes.geometry):
            for n, geom_to in enumerate(edge_nodes.geometry):
                shortest_path[i, n] = self._edges.get_dijkstra_path_length(geom_from, geom_to)

        return shortest_path

    def plot(self, edge_color="blue", node_color="red"):
        """

        :param edge_color:
        :param node_color:
        :return:
        """
        self.edges.plot(layer_color=edge_color)
        self.nodes.plot(layer_color=node_color)

    @property
    def graph(self):
        if self._graph is None:
            raise NetworkError("Corresponding graph has not been built")
        else:
            return self._graph


class RoadNetwork(Network):
    """ Road network class

    Road network is basically a multi-directed graph with methods
    for computing fuel consumption and travel time of corresponding
    vehicles.
    """

    roads = protected_property("edges")

    def __init__(self, roads, nodes, *args, **kwargs):
        """

        :param roads: Road instance
        :param nodes: road nodes
        """
        check_type(roads, Road, nodes, RoadNode)

        super().__init__(roads, nodes, *args, **kwargs)

    def build_graph(self, weight_one_way=None, weight_return=None):
        """ Build corresponding multi-directed graph

        :param weight_one_way: array of weight values for edge in one_way direction
        :param weight_return: array of weight values for edge in reverse direction
        :return:
        """
        if weight_one_way is None:
            weight_one_way = self._edges.length
        if weight_return is None:
            weight_return = self._edges.length

        if len(weight_one_way) != len(self._edges) or len(weight_return) != len(self._edges):
            raise NetworkError("Input argument(s) must have the same length as network edges")

        weight, from_node, to_node = [], [], []
        for idx, coords in enumerate(self._edges.from_to):
            if self._edges.direction[idx] != "reverse":
                weight.append(weight_one_way[idx])
                from_node.append(coords[0])
                to_node.append(coords[1])
            if self._edges.direction[idx] != "one-way":
                weight.append(weight_return[idx])
                from_node.append(coords[1])
                to_node.append(coords[0])

        # Create multi-directed graph
        self._graph = nx.MultiDiGraph()
        self._graph.add_weighted_edges_from([(from_n, to_n, w) for from_n, to_n, w in zip(from_node, to_node, weight)])

        return self

    def fuel_consumption(self, gross_hp, vehicle_weight, vehicle_frontal_area=7.92, engine_efficiency=0.4,
                         fuel_energy_density=35, uphill_hp=0.8, downhill_hp=0.6, drag_resistance=0.35,
                         mass_correction_factor=1.05, acceleration_rate=1.5 * 0.3048,
                         deceleration_rate=-9.5 * 0.3048, rho_air=1.225):
        """ Compute fuel consumption on road segments

        :param vehicle_weight:
        :param gross_hp:
        :param vehicle_frontal_area:
        :param engine_efficiency:
        :param fuel_energy_density: fuel efficiency as L/MJ
        :param uphill_hp:
        :param downhill_hp:
        :param drag_resistance:
        :param mass_correction_factor:
        :param acceleration_rate:
        :param deceleration_rate:
        :param rho_air: air density
        :return:
        """

        slope = [self.roads.slope_of_geometry(i, slope_format="degree") for i in range(len(self.roads))]
        r_curvature = [self.roads.radius_of_curvature_of_geometry(i) for i in range(len(self.roads))]
        road_length = [self.roads.length_xyz_of_geometry(i) for i in range(len(self.roads))]

        # Maximum limited speed
        v_max = self._get_max_limited_speed(slope, r_curvature, vehicle_weight, gross_hp, uphill_hp, downhill_hp)

        # Maximum speed at intersection
        v_in_max, v_out_max = self._get_velocity_at_intersection()

        # Fuel demand
        fuel_demand = {'one-way': [], 'reverse': []}

        for n in range(len(self.roads)):

            # Travel time and distance of acceleration
            t_time_one_way, d_a_one_way, _ = get_travel_time_and_distance_of_acceleration(
                v_max["one-way"][n], road_length[n], v_in_max[n], v_out_max[n], acceleration_rate, deceleration_rate)
            t_time_reverse, d_a_reverse, _ = get_travel_time_and_distance_of_acceleration(
                v_max["reverse"][n], road_length[n], v_out_max[n], v_in_max[n], acceleration_rate, deceleration_rate)

            # Travel time (for mean velocity over road segment)
            v_mean_one_way = road_length[n] / t_time_one_way
            v_mean_reverse = road_length[n] / t_time_reverse

            # Energy demand
            u_r = self.roads.rolling_coefficient[n] * vehicle_weight * 9.81 * np.cos(slope[n] * np.pi / 180) * \
                road_length[n]
            u_a_one_way = 0.5 * rho_air * vehicle_frontal_area * drag_resistance * v_mean_one_way ** 2 * road_length[n]
            u_a_reverse = 0.5 * rho_air * vehicle_frontal_area * drag_resistance * v_mean_reverse ** 2 * road_length[n]
            u_i_one_way = mass_correction_factor * vehicle_weight * acceleration_rate * d_a_one_way
            u_i_reverse = mass_correction_factor * vehicle_weight * acceleration_rate * d_a_reverse
            u_g_one_way = vehicle_weight * 9.81 * np.sin(slope[n] * np.pi / 180) * road_length[n]
            u_g_reverse = vehicle_weight * 9.81 * np.sin(-slope[n] * np.pi / 180) * road_length[n]

            fuel_demand["one-way"].append(np.maximum(0, (u_r + u_a_one_way + u_i_one_way + u_g_one_way) * 1e-6 / (
                    fuel_energy_density * engine_efficiency)))
            fuel_demand["reverse"].append(np.maximum(0, (u_r + u_a_reverse + u_i_reverse + u_g_reverse) * 1e-6 / (
                    fuel_energy_density * engine_efficiency)))

        return fuel_demand

    def travel_time(self, gross_hp, vehicle_weight, acceleration_rate=1.5 * 0.3048, deceleration_rate=-9.5 * 0.3048,
                    uphill_hp=0.8, downhill_hp=0.6, time_format='h'):
        """ Compute travel time for each road segment

        Compute travel time for each road element according
        to given parameters
        :param gross_hp: gross horse power of the vehicle
        :param vehicle_weight: weight of the vehicle
        :param acceleration_rate: positive acceleration value
        :param deceleration_rate: negative acceleration value (deceleration)
        :param uphill_hp: available horsepower on uphill road (%)
        :param downhill_hp: available horsepower on downhill road (%)
        :param time_format: format of output time (seconds, minutes, hours)
        :return:
        """
        travel_time = {'one-way': [], 'reverse': []}
        slope = [self.roads.slope_of_geometry(i, slope_format="degree") for i in range(len(self.roads))]
        r_curvature = [self.roads.radius_of_curvature_of_geometry(i) for i in range(len(self.roads))]
        road_length = [self.roads.length_xyz_of_geometry(i) for i in range(len(self.roads))]

        # Maximum limited speed
        v_max = self._get_max_limited_speed(slope, r_curvature, vehicle_weight, gross_hp, uphill_hp, downhill_hp)

        # Maximum speed at intersection
        v_in_max, v_out_max = self._get_velocity_at_intersection()

        for v, d, v_in, v_out in zip(v_max["one-way"], road_length, v_in_max, v_out_max):
            time, _, _ = get_travel_time_and_distance_of_acceleration(v, d, v_in, v_out, acceleration_rate,
                                                                      deceleration_rate)
            travel_time["one-way"].append(TIME_FORMAT[time_format] * time)

        for v, d, v_in, v_out in zip(v_max["reverse"], road_length, v_in_max, v_out_max):
            time, _, _ = get_travel_time_and_distance_of_acceleration(v[::-1], d[::-1], v_out, v_in, acceleration_rate,
                                                                      deceleration_rate)
            travel_time["reverse"].append(TIME_FORMAT[time_format] * time)

        return travel_time

    def velocity(self, gross_hp, vehicle_weight, acceleration_rate=1.5 * 0.3048, deceleration_rate=-9.5 * 0.3048,
                 uphill_hp=0.8, downhill_hp=0.6):
        """ Compute velocity for each road segment

        :param gross_hp:
        :param vehicle_weight:
        :param acceleration_rate:
        :param deceleration_rate:
        :param uphill_hp:
        :param downhill_hp:
        :return:
        """
        velocity = {'one-way': [], 'reverse': []}
        slope = [self.roads.slope_of_geometry(i, slope_format="degree") for i in range(len(self.roads))]
        r_curvature = [self.roads.radius_of_curvature_of_geometry(i) for i in range(len(self.roads))]
        road_length = [self.roads.length_xyz_of_geometry(i) for i in range(len(self.roads))]

        # Maximum limited speed
        v_max = self._get_max_limited_speed(slope, r_curvature, vehicle_weight, gross_hp, uphill_hp, downhill_hp)

        # Maximum speed at intersection
        v_in_max, v_out_max = self._get_velocity_at_intersection()

        for v, d, v_in, v_out in zip(v_max["one-way"], road_length, v_in_max, v_out_max):
            _, _, speed = get_travel_time_and_distance_of_acceleration(v, d, v_in, v_out, acceleration_rate,
                                                                       deceleration_rate)
            velocity["one-way"].append(speed)

        for v, d, v_in, v_out in zip(v_max["reverse"], road_length, v_in_max, v_out_max):
            _, _, speed = get_travel_time_and_distance_of_acceleration(v[::-1], d[::-1], v_out, v_in,
                                                                       acceleration_rate, deceleration_rate)
            velocity["reverse"].append(speed)

        velocity["v_max_one_way"] = v_max["one-way"]
        velocity["v_max_reverse"] = v_max["reverse"]
        velocity["v_slope_one_way"] = v_max["slope_one_way"]
        velocity["v_slope_reverse"] = v_max["slope_reverse"]
        velocity["v_curvature"] = v_max["curvature"]

        return velocity

    def _get_velocity_at_intersection(self):
        """ Velocity in crossing intersections

        Define maximum allowed entering and exiting velocities for each road segment
        :return:
        """
        node_coords = [(x, y) for x, y in zip(self.nodes.geometry.x, self.nodes.geometry.y)]
        v_in = np.full(len(self.roads), 0)
        v_out = np.full(len(self.roads), 0)

        for from_to in self.roads.from_to:
            v_in[node_coords.index(from_to[0])] = self.nodes.max_speed[node_coords.index(from_to[0])]
            v_out[node_coords.index(from_to[1])] = self.nodes.max_speed[node_coords.index(from_to[1])]

        return v_in, v_out

    #################
    # Private methods

    def _get_max_limited_speed(self, slope, r_curvature, vehicle_weight, gross_hp, uphill_hp, downhill_hp):
        """ Get maximum limited speed on road segments

        :param slope:
        :param r_curvature:
        :param vehicle_weight:
        :param gross_hp:
        :param uphill_hp:
        :param downhill_hp:
        :return:
        """
        v_max = dict(slope_one_way=[], slope_reverse=[], curvature=[])

        # Maximum speed due to slope (1 mechanical hp = 745.699872 W)
        ehp_uphill = gross_hp * uphill_hp * 745.699872
        ehp_downhill = gross_hp * downhill_hp * 745.699872
        for n in range(len(self.roads)):
            v_one_way = np.zeros(len(slope[n]))
            v_reverse = np.zeros(len(slope[n]))
            grade_resistance = 9.81 * vehicle_weight * np.sin(np.fabs(slope[n]) * np.pi / 180)
            rolling_resistance = 9.81 * self.roads.rolling_coefficient[n] * vehicle_weight * np.cos(slope[n] * np.pi /
                                                                                                    180)
            v_one_way[slope[n] < 0] = ehp_downhill / np.maximum((grade_resistance[slope[n] < 0] - rolling_resistance[
                slope[n] < 0]), 0)
            v_one_way[slope[n] >= 0] = ehp_uphill / (grade_resistance[slope[n] >= 0] + rolling_resistance[slope[n] >=
                                                                                                          0])
            v_reverse[slope[n] > 0] = ehp_downhill / np.maximum((grade_resistance[slope[n] > 0] - rolling_resistance[
                slope[n] > 0]), 0)
            v_reverse[slope[n] <= 0] = ehp_uphill / (grade_resistance[slope[n] <= 0] + rolling_resistance[slope[n] <=
                                                                                                          0])
            v_max["slope_one_way"].append(v_one_way)
            v_max["slope_reverse"].append(v_reverse)
            v_max["curvature"].append((self.roads.rollover_criterion[n] * r_curvature[n] * 9.81) ** 0.5)

        # Get maximum limiting speed, i.e. minimum among all previous values
        v_max["one-way"] = [np.minimum(np.minimum(v_r, v_s), v_limit) for v_r, v_s, v_limit in
                            zip(v_max["curvature"], v_max["slope_one_way"], self.roads.max_speed)]
        v_max["reverse"] = [np.minimum(np.minimum(v_r, v_s), v_limit) for v_r, v_s, v_limit in
                            zip(v_max["curvature"], v_max["slope_reverse"], self.roads.max_speed)]

        return v_max


@jit((float64[:], float64[:], float64, float64, float64, float64), cache=True, nopython=True)
def get_travel_time_and_distance_of_acceleration(v_max, road_segment_length, v_in_max, v_out_max, a_1, a_2):
    """ Get travel time on road segment

    :param v_max: maximum limited speed on road segment
    :param road_segment_length: length of road segment
    :param v_in_max:
    :param v_out_max:
    :param a_1: acceleration rate
    :param a_2: deceleration rate
    :return:
    """

    v = np.concatenate((np.array([v_in_max]), v_max[1:], np.array([v_out_max])))
    t_time = np.zeros(len(v_max))
    d_a = np.zeros(len(v_max))  # Distance of acceleration

    tol = 0.01  # Tolerance for comparing d <= s --> d must be <= s + tolerance (In order to avoid too much
    # backward in the while loop, as well as floating errors where d is not exactly equal to s at 1e-10)
    n = 0
    while n <= len(v_max) - 1:
        v_in = v[n]
        v_fn = v[n + 1]
        s = road_segment_length[n]
        v_m = v_max[n]

        d_1 = (v_m ** 2 - v_in ** 2) / (2 * a_1)
        d_2 = (v_fn ** 2 - v_m ** 2) / (2 * a_2)
        if v_m > v_in and v_m > v_fn:
            if v_fn >= v_in:
                d = (v_fn ** 2 - v_in ** 2) / (2 * a_1)
            else:
                d = (v_fn ** 2 - v_in ** 2) / (2 * a_2)
            if d_1 + d_2 <= s:
                t_time[n] = (v_m - v_in) / a_1 + (v_fn - v_m) / a_2 + (s - (d_1 + d_2)) / v_m
                d_a[n] = d_1
                v[n + 1] = v_fn
                n += 1
            else:
                if d <= s + tol:
                    v_min = ((2 * s * a_1 * a_2 + a_2 * v_in ** 2 - a_1 * v_fn ** 2) / (a_2 - a_1)) ** 0.5
                    t_time[n] = (v_min - v_in) / a_1 + (v_fn - v_min) / a_2
                    d_a[n] = (v_min ** 2 - v_in ** 2) / (2 * a_1)
                    v[n + 1] = v_fn
                    n += 1
                else:
                    if v_fn >= v_in:
                        v_front = (v_in ** 2 + 2 * a_1 * s) ** 0.5
                        t_time[n] = (v_front - v_in) / a_1
                        d_a[n] = s
                        v[n + 1] = v_front
                        n += 1
                    else:
                        v[n] = (v_fn ** 2 - 2 * a_2 * s) ** 0.5
                        n -= 1 if n > 0 else 0
        elif v_fn < v_m <= v_in:
            if d_2 <= s + tol:
                t_time[n] = (v_fn - v_m) / a_2 + (s - d_2) / v_m
                v[n + 1] = v_fn
                n += 1
            else:
                v[n] = (v_fn ** 2 - 2 * a_2 * s) ** 0.5
                n -= 1 if n > 0 else 0
        elif v_in < v_m <= v_fn:
            if d_1 <= s:
                t_time[n] = (v_m - v_in) / a_1 + (s - d_1) / v_m
                d_a[n] = d_1
                v[n + 1] = v_m
            else:
                v_front = (v_in ** 2 + 2 * a_1 * s) ** 0.5
                t_time[n] = (v_front - v_in) / a_1
                d_a[n] = s
                v[n + 1] = v_front
            n += 1
        elif v_m <= v_in and v_m <= v_fn:
            t_time[n] = s / v_m
            v[n + 1] = v_m
            n += 1

    return t_time, d_a, v



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

# __all__ = ["compute_surface"]
# __version__ = '0.1'
__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2018, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'

import numpy as np
from numpy import sin, pi, sqrt, abs
from utils.check.type import check_type

from gistools.coordinates import Ellipsoid


def compute_surface(lb_pixel, rb_pixel, ub_pixel, bb_pixel, geo_type: str, ellipsoid: Ellipsoid):
    """ Compute surface of earth pixels

    Combine surface element in spherical coordinates
    and local spherical approximation of the given
    ellipsoid to compute pixel surface. Or return
    surface from evenly spaced pixels from equiarea
    map projection.
    :param lb_pixel: left border pixel coordinate
    :param rb_pixel: right border pixel coordinate
    :param ub_pixel: upper border pixel coordinate
    :param bb_pixel: bottom border pixel coordinate
    :param geo_type: geo projection ("latlon" or "equal")
    :param ellipsoid: ellipsoid model
    :return:
    """

    latitude = (ub_pixel + bb_pixel) / 2

    def ellipsoid_normal():
        return ellipsoid.a / sqrt(1 - (ellipsoid.e ** 2) * (sin(latitude * pi / 180)) ** 2)

    def local_sphere_radius():

        p = ellipsoid.a * (1 - ellipsoid.e ** 2) / (1 - ellipsoid.e ** 2 *
                                                    sin(latitude * pi/180) ** 2) ** (3/2)
        return sqrt(ellipsoid_normal() * p) * 10 ** -3

    check_type(geo_type, str, ellipsoid, Ellipsoid)

    if geo_type == "latlon":
        surface = local_sphere_radius() ** 2 * (sin(ub_pixel * pi / 180) - sin(bb_pixel * pi / 180)) *\
                  (rb_pixel * pi / 180 - lb_pixel * pi / 180)
    elif geo_type == "equal":
        surface = abs(ub_pixel - bb_pixel) * abs(rb_pixel - lb_pixel)
    else:
        raise ValueError("Projection {} has not been defined yet".format(geo_type))

    return surface


def pixel_overlap_area(centroid1, centroid2, res1, res2, geoproj: str, ellipsoid: Ellipsoid):
    """ Retrieve overlay area between pixels

    Compute overlap area between pixels
    (thanks to https://stackoverflow.com/questions/9324339/how-much-do-two-rectangles-overlap for inspiration)
    :param centroid1:
    :param centroid2:
    :param res1:
    :param res2:
    :param geoproj:
    :param ellipsoid:
    :return:
    """

    def get_pixel_borders(centroid, res):
        return centroid[0] - res/2, centroid[0] + res/2, centroid[1] + res/2, centroid[1] - res/2

    lb_pixel1, rb_pixel1, ub_pixel1, bb_pixel1 = get_pixel_borders(centroid1, res1)
    lb_pixel2, rb_pixel2, ub_pixel2, bb_pixel2 = get_pixel_borders(centroid2, res2)

    # Overlapping area
    lb_area = np.maximum(lb_pixel1, lb_pixel2)
    rb_area = np.minimum(rb_pixel1, rb_pixel2)
    ub_area = np.minimum(ub_pixel1, ub_pixel2)
    bb_area = np.maximum(bb_pixel1, bb_pixel2)

    # Overlapping (west-east and south-north)
    x_overlap = np.maximum(0, rb_area - lb_area)
    y_overlap = np.maximum(0, ub_area - bb_area)

    overlapping_area = compute_surface(lb_area, rb_area, ub_area, bb_area, geoproj, ellipsoid)
    overlapping_area[np.isclose(x_overlap, 0) | np.isclose(y_overlap, 0)] = 0

    return overlapping_area



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

# __all__ = []
# __version__ = '0.1'

import numpy as np
import networkx as nx

from math import sqrt as msqrt

from shapely.errors import TopologicalError
from shapely.geometry import MultiPolygon, GeometryCollection, Polygon, box, LineString, \
    Point, MultiLineString, JOIN_STYLE
from shapely.ops import cascaded_union, linemerge, unary_union, transform
from utils.check.type import is_iterable, type_assert

from gistools.coordinates import r_tree_idx
from gistools.graph import part_graph

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2018, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


def add_points_to_line(line, threshold):
    """ Add point coordinates to line geometry

    :param line:
    :param threshold:
    :return:
    """
    return linemerge(cut_(line, threshold))


def aggregate_partitions(polygons, weights, nparts, division, weight_attr, split, recursive, **metis_options):
    """ Aggregate polygons into partitions

    :param polygons: polygons to aggregate
    :param weights: polygons' corresponding weight
    :param nparts: number of partitions
    :param division: list of final relative weights of each partition
    :param weight_attr:
    :param split:
    :param recursive:
    :param metis_options:
    :return:
    """
    if "contig" not in metis_options.keys():
        metis_options["contig"] = False
    graph = polygon_collection_to_graph(polygons, weights, split, metis_options["contig"], weight_attr)
    tpweights = [(d,) for d in division]
    partition = part_graph(graph, nparts, weight_attr, tpweights, recursive, **metis_options)

    # Return unions of polygons belonging to each part (no multi-polygons)
    return explode([no_artifact_unary_union([polygons[n] for n in part]) for part in partition])


def area_partition_polygon(polygon, unit_area, disaggregation_factor, precision, recursive, split, **metis_options):
    """ Partition polygon into a subset of polygons of equal area

    :param polygon: polygon intended to be partitioned
    :param unit_area: area of a sub-polygon
    :param disaggregation_factor: factor use to discretize polygons before aggregation
    :param recursive: k-way or recursive method for partitioning
    :param precision: metric precision for sub-polygon area
    :param split: function used to split polygon into smaller unit blocks
    :param metis_options: specific METIS options (see METIS manual)
    :return:
    """
    nparts = int(polygon.area/unit_area)

    if nparts <= 1 and (polygon.area - unit_area) < unit_area/disaggregation_factor:
        return [polygon]

    # Split polygon into sub-elements
    split_poly = split_polygon(polygon, split, unit_area/disaggregation_factor, get_explode=True)

    division = [unit_area/polygon.area] * nparts
    if polygon.area % unit_area != 0:  # and (polygon.area - nparts * unit_area) >= unit_area/disaggregation_factor:
        division += [(polygon.area - nparts * unit_area)/polygon.area]
        nparts += 1

    area = [int(poly.area / precision) for poly in split_poly]

    return aggregate_partitions(split_poly, area, nparts, division, "area", split, recursive, **metis_options)


def centroid(point_collection):
    """ Retrieve centroid of multiple points

    :param point_collection:
    :return:
    """
    x_centroid = np.mean([pt.x for pt in point_collection])
    y_centroid = np.mean([pt.y for pt in point_collection])

    return Point([x_centroid, y_centroid])


def connect_lines_to_point(line_collection, point):
    """ Connect a set of lines to some point

    :param line_collection:
    :param point:
    :return:
    """
    new_line_collection = []
    for line in line_collection:
        if Point(line.coords[0]).distance(point) < Point(line.coords[-1]).distance(point):
            new_line_collection.append(LineString(point.coords[:] + line.coords[:]))
        else:
            new_line_collection.append(LineString(line.coords[:] + point.coords[:]))

    return new_line_collection


def cut(line, threshold, count=0):
    """ Cut a line in segments

    Cut a line in segments whose length
    is below a threshold value. This method
    is more randomless regarding the final
    size of the line segments. See 'cut_'
    function for more accuracy
    :param line:
    :param threshold:
    :param count:
    :return:
    """
    result = []
    if threshold < 0 or threshold >= line.length or count == 250:
        return [line]
    # Recursively cut line in 2 at midpoint
    p = line.interpolate(0.5, normalized=True)
    split_line = cut_at_point(line, p)
    for sub_line in split_line:
        result.extend(cut(sub_line, threshold, count + 1))

    return result


def cut_(line, threshold):
    """ Cut a line in segments (method 2)

    This method cuts a line in as many segments as necessary,
    depending on the given threshold. For instance, a line
    of 105m will be cut into 10 pieces of 10m + 1 piece of 5m
    if threshold=10
    :param line: LineString
    :param threshold: minimum sub line piece size
    :return:
    """
    if threshold < 0 or threshold >= line.length:
        return [line]

    result = []

    while "It remains line to cut":
        split_line = cut_at_distance(line, threshold/line.length, normalized=True)
        result.append(split_line[0])

        if split_line[1].length > threshold:
            line = split_line[1]
        else:
            result.append(split_line[1])
            break

    return result


def cut_at_distance(line, distance, normalized=False):
    """ Cut line at given distance from starting point

    :param line:
    :param distance:
    :param normalized:
    :return:
    """
    if normalized:
        length = 1
    else:
        length = line.length

    if distance <= 0.0 or distance >= length:
        return [line]
    coords = list(line.coords)
    for i, p in enumerate(coords):
        pd = line.project(Point(p), normalized=normalized)
        if pd == distance:
            return [LineString(coords[:i+1]), LineString(coords[i:])]
        elif pd > distance:
            cp = line.interpolate(distance, normalized=normalized)
            try:
                return [LineString(coords[:i] + [(cp.x, cp.y)]), LineString([(cp.x, cp.y)] + coords[i:])]
            except ValueError:
                return [LineString(coords[:i] + [(cp.x, cp.y, cp.z)]), LineString([(cp.x, cp.y, cp.z)] + coords[i:])]


def cut_at_point(line, point):
    """ Cut line at point

    Cut line at point, which can be within
    or without the geometry
    :param line:
    :param point:
    :return:
    """
    d = line.project(point)
    return cut_at_distance(line, d)


def cut_at_points(line, points):
    """ Cut line at multiple points

    :param line:
    :param points:
    :return:
    """
    cut_line = []
    distance = [line.project(point) for point in points]
    sorted_points = [point for _, point in sorted(zip(distance, points))]

    for idx, point in enumerate(sorted_points):
        cut_line.extend(cut_at_point(line, point))
        if idx < len(sorted_points) - 1:
            line = cut_line.pop()

    return cut_line


def dissolve(geometry_collection):
    """ Recursively join contiguous geometries in collection

    :param geometry_collection:
    :return:
    """
    if not is_iterable(geometry_collection):
        raise TypeError("Input must be a collection but is '{}'".format(type(geometry_collection)))

    while "There is still geometries to aggregate":

        joint = []
        idx = r_tree_idx(geometry_collection)
        geom_idx = []
        increment = 0

        while len(geom_idx) < len(geometry_collection):

            if increment not in geom_idx:
                geom = geometry_collection[increment]
                union_idx, union = intersecting_features(geom, geometry_collection, idx)

                if len(union) > 0:
                    joint.append(cascaded_union(union))

                for ix in union_idx:
                    idx.delete(ix, geometry_collection[ix].bounds)

                geom_idx.extend(union_idx)

            increment += 1

        if len(joint) < len(geometry_collection):
            geometry_collection = joint
        else:
            break

    return joint


def explode(geometry_collection):
    """ Convert multi-part geometry collection into single-part

    :param geometry_collection: valid geometry collection
    :return:
    """
    single = []
    if not is_iterable(geometry_collection):
        geometry_collection = [geometry_collection]

    for geom in geometry_collection:
        try:
            single.extend(geom)
        except TypeError:
            single.append(geom)

    return single


def fishnet(polygon, threshold):
    """ Intersect polygon with a regular grid or "fishnet"

    :param polygon:
    :param threshold:
    :return:
    """
    return polygon_to_mesh(polygon, threshold, mesh)


def hexana(polygon, threshold):
    """ Split a polygon using a honeycomb grid

    :param polygon: original polygon to split
    :param threshold: unit hexagon surface
    :return: list of polygons
    """
    return polygon_to_mesh(polygon, threshold, honeycomb)


# Thanks to https://gist.github.com/urschrei/17cf0be92ca90a244a91
def honeycomb(startx, starty, endx, endy, radius=None, area=None):
    """
    Calculate a grid of hexagon coordinates of the given radius
    given lower-left and upper-right coordinates
    Returns a list of lists containing 6 tuples of x, y point coordinates
    These can be used to construct valid regular hexagonal polygons

    - update 04/23/2019:
        * can give either radius or area of unit hexagon
        * return a list of shapely Polygon

    You will probably want to use projected coordinates for this
    """
    if not radius:
        radius = msqrt(area / (2*msqrt(3)))

    # calculate side length given radius
    sl = (2 * radius) * np.tan(np.pi / 6)
    # calculate radius for a given side-length
    # (a * (math.cos(math.pi / 6) / math.sin(math.pi / 6)) / 2)
    # see http://www.calculatorsoup.com/calculators/geometry-plane/polygon.php

    # calculate coordinates of the hexagon points
    # sin(30)
    p = sl * 0.5
    b = sl * np.cos(np.radians(30))
    w = b * 2
    h = 2 * sl

    # offset start and end coordinates by hex widths and heights to guarantee coverage
    startx = startx - w
    starty = starty - h
    endx = endx + w
    endy = endy + h
    origx = startx

    # offsets for moving along and up rows
    xoffset = b
    yoffset = 3 * p

    polygons = []
    row = 1
    counter = 0

    while starty < endy:
        if row % 2 == 0:
            startx = origx + xoffset
        else:
            startx = origx
        while startx < endx:
            p1x = startx
            p1y = starty + p
            p2x = startx
            p2y = starty + (3 * p)
            p3x = startx + b
            p3y = starty + h
            p4x = startx + w
            p4y = starty + (3 * p)
            p5x = startx + w
            p5y = starty + p
            p6x = startx + b
            p6y = starty
            poly = [
                (p1x, p1y),
                (p2x, p2y),
                (p3x, p3y),
                (p4x, p4y),
                (p5x, p5y),
                (p6x, p6y),
                (p1x, p1y)]
            polygons.append(Polygon(poly))
            counter += 1
            startx += w
        starty += yoffset
        row += 1

    return polygons


def intersecting_features(geometry, geometry_collection, r_tree=None):
    """ Return list of geometries intersecting with given geometry

    :param geometry:
    :param geometry_collection:
    :param r_tree: rtree index corresponding to geometry collection
    :return:
    """
    is_intersecting = intersects(geometry, geometry_collection, r_tree)
    return [i for i in range(len(geometry_collection)) if is_intersecting[i]], \
           [geom for i, geom in enumerate(geometry_collection) if is_intersecting[i]]


def intersects(geometry, geometry_collection, r_tree=None):
    """ Return if geometry intersects with geometries of collection

    Use this function with large geometry collections
    :param geometry:
    :param geometry_collection:
    :param r_tree:
    :return: list of boolean of length = length(geometry_collection)
    """
    # Use Rtree to speed up !
    if r_tree is None:
        r_tree = r_tree_idx(geometry_collection)

    list_of_intersecting_features = list(r_tree.intersection(geometry.bounds))

    return [False if f not in list_of_intersecting_features else geometry.intersects(geometry_collection[f]) for f in
            range(len(geometry_collection))]


def is_in_collection(geometry, geometry_collection, r_tree):
    """ Test if geometry is present in collection (using shapely 'equals' method)

    :param geometry:
    :param geometry_collection:
    :param r_tree:
    :return:
    """
    _, list_of_intersecting_features = intersecting_features(geometry, geometry_collection, r_tree)
    for geom in list_of_intersecting_features:
        if geometry.equals(geom):
            return True

    return False


def is_line_connected_to(line, geometry_collection):
    """ Is line connected to one of the geometries in collection ?

    :param line:
    :param geometry_collection:
    :return:
    """

    return [other.intersects(Point(line.coords[0])) for other in geometry_collection], [other.intersects(Point(
        line.coords[-1])) for other in geometry_collection]


def katana(polygon, threshold, count=0):
    """ Split a polygon

    See https://snorfalorpagus.net/blog/2016/03/13/splitting-large-polygons-for-faster-intersections/

    Copyright (c) 2016, Joshua Arnott

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
    following conditions are met:

    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
    disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
    WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
    GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    :param polygon: Shapely polygon
    :param threshold:
    :param count:
    :return:
    """
    if count == 0:
        if not polygon.is_valid:
            polygon = polygon.buffer(0, 0)

    result = []
    width = polygon.bounds[2] - polygon.bounds[0]
    height = polygon.bounds[3] - polygon.bounds[1]
    if width * height <= threshold or count == 250:
        return [polygon]
    if height >= width:
        a = box(polygon.bounds[0], polygon.bounds[1], polygon.bounds[2], polygon.bounds[1] + height/2)
        b = box(polygon.bounds[0], polygon.bounds[1] + height/2, polygon.bounds[2], polygon.bounds[3])
    else:
        a = box(polygon.bounds[0], polygon.bounds[1], polygon.bounds[0] + width/2, polygon.bounds[3])
        b = box(polygon.bounds[0] + width/2, polygon.bounds[1], polygon.bounds[2], polygon.bounds[3])

    for sword in (a, b,):
        split_poly = polygon.intersection(sword)
        if not isinstance(split_poly, GeometryCollection):
            split_poly = [split_poly]
        for sub_poly in split_poly:
            if isinstance(sub_poly, (Polygon, MultiPolygon)):
                result.extend(katana(sub_poly, threshold, count+1))

    return result


def katana_centroid(polygon, threshold, count=0):
    """ Split a polygon in equal areas

    Thanks to https://snorfalorpagus.net/blog/2016/03/13/splitting-large-polygons-for-faster-intersections/ and
    Daniel Harasty in http://community-gispython-org-community-projects.955323.n3.nabble.com/Community-Spliting-a
    -polygon- into-two-polygons-with-the-same-area-td4024026.html#a4024033, we merge here both approaches to split a
    polygon into a number of sub-polygons of almost equal areas.
    :param polygon: Shapely polygon
    :param threshold:
    :param count:
    :return:
    """
    if count == 0:
        if not polygon.is_valid:
            polygon = polygon.buffer(0, 0)

    result = []
    width = polygon.bounds[2] - polygon.bounds[0]
    height = polygon.bounds[3] - polygon.bounds[1]
    if width * height <= threshold or count == 250:
        return [polygon]
    if height >= width:
        a = box(polygon.bounds[0], polygon.bounds[1], polygon.bounds[2], polygon.centroid.y)
        b = box(polygon.bounds[0], polygon.centroid.y, polygon.bounds[2], polygon.bounds[3])
    else:
        a = box(polygon.bounds[0], polygon.bounds[1], polygon.centroid.x, polygon.bounds[3])
        b = box(polygon.centroid.x, polygon.bounds[1], polygon.bounds[2], polygon.bounds[3])

    for sword in (a, b,):
        split_poly = polygon.intersection(sword)
        if not isinstance(split_poly, GeometryCollection):
            split_poly = [split_poly]
        for sub_poly in split_poly:
            if isinstance(sub_poly, (Polygon, MultiPolygon)):
                result.extend(katana_centroid(sub_poly, threshold, count+1))

    return result


def length_of_segments(line):
    """ Retrieve segment length in line

    :param line:
    :return:
    """
    return np.diff([line.project(Point(p)) for p in line.coords])


def mask(polygon_collection, mask_collection, fast_intersection_surface):
    """ Geometry mask

    :param polygon_collection:
    :param mask_collection:
    :param fast_intersection_surface:
    :return:
    """

    # Retrieve base layer and mask geometry, split it for faster intersection
    # and explode it (to be sure there is no multi-parts)
    geometry = split_polygon_collection(polygon_collection, fast_intersection_surface, get_explode=True)
    mask_geometry = split_polygon_collection(mask_collection, fast_intersection_surface, get_explode=True)

    # Use Rtree to speed up !
    idx = r_tree_idx(mask_geometry)

    # 0. Initialization
    result = []

    for geom in geometry:
        list_of_intersecting_mask = list(idx.intersection(geom.bounds))
        within = [geom.within(mask_geometry[n]) for n in list_of_intersecting_mask]
        if not any(within):
            is_intersecting = [geom.intersects(mask_geometry[n]) for n in list_of_intersecting_mask]
            if any(is_intersecting):
                difference = geom.difference(cascaded_union([mask_geometry[n] for n in list_of_intersecting_mask]))
                if not difference.is_empty:
                    result.append(difference)
            else:
                result.append(geom)

    # Multi to single + dissolve coincident polygons
    result = explode(result)
    result = [no_artifact_unary_union(poly) for poly in dissolve(result)]

    return result


def merge(line_collection):
    """ Merge connected lines

    :param line_collection:
    :return:
    """
    # Merge MultiLinestring objects returned by the "join" function
    merged_line = [linemerge(line) if isinstance(line, MultiLineString) else line for line in dissolve(line_collection)]

    # Keep only single parts
    return explode(merged_line)


def mesh(startx, starty, endx, endy, side=None, area=None):
    """ Compute a mesh grid

    :param startx:
    :param starty:
    :param endx:
    :param endy:
    :param side:
    :param area:
    :return:
    """
    if not side:
        side = msqrt(area)

    startx = startx - side/2
    starty = starty - side/2
    endx = endx + side/2
    endy = endy + side/2
    origx = startx

    polygons = []
    while starty < endy:
        startx = origx
        while startx < endx:
            poly = [
                (startx, starty),
                (startx, starty + side),
                (startx + side, starty + side),
                (startx + side, starty)]
            polygons.append(Polygon(poly))
            startx += side
        starty += side

    return polygons


def nearest_feature(geometry, geometry_collection, r_tree=None):
    """ Return nearest feature from geometry collection to given geometry

    If some of the geometries intersect, the nearest feature is the one whose centroid is the closest to the centroid
    of the given geometry (but distance remains 0)
    :param geometry:
    :param geometry_collection:
    :param r_tree: rtree index corresponding to geometry collection
    :return: nearest feature index and corresponding distance
    """
    # Use Rtree to speed up !
    if r_tree is None:
        r_tree = r_tree_idx(geometry_collection)

    # Look if some geometries intersect
    list_of_intersecting_features, _ = intersecting_features(geometry, geometry_collection, r_tree)

    if list_of_intersecting_features:
        distance = [geometry.centroid.distance(geometry_collection[n].centroid) for n in list_of_intersecting_features]
        return list_of_intersecting_features[np.argmin(distance)], 0
    else:
        list_of_nearest_features = list(r_tree.nearest(geometry.bounds, 1))
        distance = [geometry.distance(geometry_collection[n]) for n in list_of_nearest_features]
        return list_of_nearest_features[np.argmin(distance)], np.min(distance)


def no_artifact_unary_union(geoms, eps=0.00001):
    """ Make unary union that does not return artifacts

    Thanks to https://gis.stackexchange.com/questions/277334/shapely-polygon-union-results-in-strange-artifacts-of
    -tiny-non-overlapping-area
    :param geoms: list of geoms to aggregate
    :param eps: buffering precision
    :return:
    """
    return unary_union(geoms).buffer(eps, 1, join_style=JOIN_STYLE.mitre).buffer(-eps, 1, join_style=JOIN_STYLE.mitre)


def overlapping_features(geometry, geometry_collection, r_tree=None):
    """ Return list of geometries overlapping with given geometry

    Overlapping geometry is either overlapping in the shapely way,
    or within or containing the other geometry
    :param geometry:
    :param geometry_collection:
    :param r_tree:
    :return:
    """
    idx, list_of_intersecting_features = intersecting_features(geometry, geometry_collection, r_tree)
    _overlaps = [[i, geom] for i, geom in zip(idx, list_of_intersecting_features) if geom.overlaps(geometry) or
                 geom.within(geometry) or geom.contains(geometry)]

    return [overlap[0] for overlap in _overlaps], [overlap[1] for overlap in _overlaps]


def overlaps(geometry, geometry_collection, r_tree=None):
    """ Return if geometry overlaps with geometries of collection

    Overlapping is regarded as any area shared by two geometries
    :param geometry:
    :param geometry_collection:
    :param r_tree:
    :return:
    """
    is_intersecting = intersects(geometry, geometry_collection, r_tree)
    return [False if not is_intersecting[i] else geom.overlaps(geometry) or geom.within(geometry) or geom.contains(
        geometry) for i, geom in enumerate(geometry_collection)]


def polygon_to_mesh(polygon, threshold, method):
    """

    :param polygon:
    :param threshold:
    :param method: {'hexana', 'fishnet'}
    :return:
    """
    grid = method(*polygon.bounds, area=threshold)
    split = []
    for unit in grid:
        if unit.within(polygon):
            split.append(unit)
        elif unit.overlaps(polygon):
            split.append(unit.intersection(polygon))

    return explode(split)


def polygon_collection_to_graph(polygon_collection, weights, split, is_contiguous, weight_attr="weight"):
    """ Convert collection of polygons to networkx graph

    Conversion of a polygon collection into a graph allows
    later graph partitioning
    :param polygon_collection:
    :param weights: weight of each polygon in collection
    :param split: split function
    :param is_contiguous: True or False (metis options)
    :param weight_attr: name of weight attribute
    :return:
    """
    if not is_iterable(polygon_collection):
        raise TypeError("Input must be a collection but is '{}'".format(type(polygon_collection)))

    if 'katana' in split.__name__:
        is_katana = True
    else:
        is_katana = False

    r_tree = r_tree_idx(polygon_collection)
    graph = nx.Graph()

    for n, polygon in enumerate(polygon_collection):
        list_of_intersecting_features, _ = intersecting_features(polygon, polygon_collection, r_tree)
        list_of_intersecting_features.remove(n)
        if list_of_intersecting_features or not is_contiguous:
            if is_katana:
                graph.add_edges_from([(n, feature) for feature in list_of_intersecting_features
                                      if not isinstance(polygon.intersection(polygon_collection[feature]), Point)])
            else:
                graph.add_edges_from([(n, feature) for feature in list_of_intersecting_features])
            graph.add_node(n, **{weight_attr: weights[n]})

    return graph


def radius_of_curvature(line, method="osculating"):
    """ Compute curvature radius of LineString

    :param line:
    :param method: method for computing radius of curvature {'circumscribe', 'osculating'}
    :return:
    """
    def norm(xx, yy):
        return np.sqrt(xx ** 2 + yy ** 2)

    def tangent_vector(xi, yi):
        return (xi[2::] - xi[:-2]) / norm(xi[2::] - xi[:-2], yi[2::] - yi[:-2]), \
               (yi[2::] - yi[:-2]) / norm(xi[2::] - xi[:-2], yi[2::] - yi[:-2])

    if method == "osculating":

        if len(line.coords) >= 3:
            x = np.array(line.coords.xy[0])
            y = np.array(line.coords.xy[1])
            xi1 = np.concatenate((x[1::], [x[-1]]))
            yi1 = np.concatenate((y[1::], [y[-1]]))
            xi_1 = np.concatenate(([x[0]], x[:-1]))
            yi_1 = np.concatenate(([y[0]], y[:-1]))

            tangent_vector_xi1, tangent_vector_yi1 = tangent_vector(xi1, yi1)
            tangent_vector_xi_1, tangent_vector_yi_1 = tangent_vector(xi_1, yi_1)

            coefficient_of_curvature = \
                norm(tangent_vector_xi1 - tangent_vector_xi_1, tangent_vector_yi1 - tangent_vector_yi_1) /\
                norm(x[2::] - x[:-2], y[2::] - y[:-2])
            coefficient_of_curvature[coefficient_of_curvature == 0] = 1e-6
            rad_of_curvature = 1 / coefficient_of_curvature
        else:
            return np.array([10000])

    elif method == "circumscribe":

        segment_length = length_of_segments(line)
        a, b = segment_length[:-1:], segment_length[1::]
        c = []
        if len(line.coords) > 3:
            length_2_by_2_start = length_of_segments(LineString(line.coords[::2]))
            length_2_by_2_end = length_of_segments(LineString(line.coords[1::2]))

            for n in range(len(length_2_by_2_end)):
                c.extend([length_2_by_2_start[n], length_2_by_2_end[n]])

            if len(length_2_by_2_start) > len(length_2_by_2_end):
                c.append(length_2_by_2_start[-1])

        elif len(line.coords) == 3:
            c = LineString(line.coords[::2]).length

        elif len(line.coords) < 3:
            return np.array([10000])

        heron = (a + b + c) * (b + c - a) * (c + a - b) * (a + b - c)
        heron[heron < 0] = 0
        divider = np.sqrt(heron)
        divider[divider == 0] = 0.1
        rad_of_curvature = a * b * c / divider

    else:
        rad_of_curvature = []

    # Return values and add replicate to beginning of array (as result of curvature computation returns an array with
    #  length = length(line.coords) - 2): return array with length = length(line.coords) - 1
    return np.concatenate(([rad_of_curvature[0]], rad_of_curvature))


def shape_factor(polygon, convex_hull):
    """ Compute shape factor of given polygon

    Compute shape factor (here, circularity) of
    a given polygon using either convex hull or not
    :param polygon:
    :param convex_hull: should convex hull be used for computing shape ? (bool)
    :return:
    """

    if convex_hull:
        return 4 * np.pi * polygon.convex_hull.area / (polygon.convex_hull.length ** 2)
    else:
        return 4 * np.pi * polygon.area / (polygon.length ** 2)


@type_assert(polygon1=(Polygon, MultiPolygon), polygon2=(Polygon, MultiPolygon), normalized=bool)
def shared_area(polygon1, polygon2, normalized=False):
    """ Get area shared by 2 polygons

    :param polygon1:
    :param polygon2:
    :param normalized:
    :return:
    """
    if not polygon1.intersects(polygon2):
        return 0
    else:
        new_poly = polygon1.intersection(polygon2)
        if normalized:
            return new_poly.area / polygon1.area
        else:
            return new_poly.area


@type_assert(polygon=(Polygon, MultiPolygon), normalized=bool)
def shared_area_among_collection(polygon: Polygon, polygon_collection, normalized: bool = False, r_tree=None):
    """ Get area shared by a polygon with polygons from a collection

    :param polygon:
    :param polygon_collection:
    :param normalized:
    :param r_tree:
    :return:
    """
    if not is_iterable(polygon_collection):
        raise TypeError("Input 2 must be a collection but is '{}'".format(type(polygon_collection)))

    poly_intersects = intersects(polygon, polygon_collection, r_tree)

    return [shared_area(polygon, poly, normalized) if poly_intersects[n] else 0 for n, poly in enumerate(
            polygon_collection)]


def split_collection(geometry_collection, threshold, method, get_explode):
    """ Split geometry collection

    :param geometry_collection:
    :param threshold:
    :param method:
    :param get_explode:
    :return:
    """
    if not is_iterable(geometry_collection):
        raise TypeError("Geometry must be a collection")

    new_collection = []

    for geom in geometry_collection:
        try:
            new_collection.extend(method(geom, threshold))
        except TopologicalError:
            new_collection.append(geom)

    if get_explode:
        new_collection = explode(new_collection)

    # Return new collection
    return new_collection


def split_line_collection(line_collection, threshold, method="cut", get_explode=False):
    """

    :param line_collection:
    :param threshold:
    :param method:
    :param get_explode:
    :return:
    """
    split_method = {'cut': cut}

    return split_collection(line_collection, threshold, split_method[method], get_explode)


def split_polygon(polygon, method, threshold, get_explode):
    """ Split polygon with respect to method

    Split polygon and return exploded (no multi part) if necessary
    :param polygon:
    :param method:
    :param threshold:
    :param get_explode: (boolean) return exploded collection
    :return:
    """
    sp_poly = method(polygon, threshold)
    if get_explode:
        return explode(sp_poly)
    else:
        return sp_poly


def split_polygon_collection(polygon_collection, threshold, method="katana", get_explode=False):
    """ Split a collection of polygons

    :param polygon_collection: collection of shapely polygons
    :param threshold: threshold surface under which no more splitting must be achieved
    :param method: method used for splitting
    :param get_explode:
    :return: new polygon collection with only Polygon geometries (no MultiPolygon geometries)
    """
    split_method = {'katana': katana, 'katana_centroid': katana_centroid}

    return split_collection(polygon_collection, threshold, split_method[method], get_explode)


def to_2d(geometry):
    """ Convert 3D geometry to 2D

    Credit to @feenster and @hunt3ri from
    https://github.com/hotosm/tasking-manager/blob/master/server/services/grid/grid_service.py
    :param geometry:
    :return:
    """
    def _to_2d(x, y, z):
        return tuple(filter(None, [x, y]))

    return transform(_to_2d, geometry)



# -*- coding: utf-8 -*-

""" Functions for computing get_horizon profiles

More detailed description.
"""
import numpy as np

from numba import jit, int64, float64

from gistools.coordinates import Ellipsoid

__version__ = '0.1'
__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2018, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


# TODO: use Numba
# TODO: use Dozier algorithm (1981)
# TODO: use algorithm from A. James Stewart (1998)


@jit((float64[:], float64[:], float64[:], float64),  nopython=True)
def dozier_queue(dem, x, y, theta):
    """

    :param dem: array-like
    :param x: array-like
    :param y: array-like
    :param theta:
    :return:
    """
    def rotate():
        return x * np.cos(theta) - y * np.sin(theta), x * np.sin(theta) + y * np.cos(theta)

    x_rotate, y_rotate = rotate()
    x_rotate = np.round_(x_rotate, 0, np.empty_like(x_rotate))
    x_rotate_values = np.unique(x_rotate)

    # profiles = [np.ndarray(shape=(1,), dtype=np.float64) for n in range(0)]
    # coord_x = [np.ndarray(shape=(1,), dtype=np.float64) for n in range(0)]
    # coord_y = [np.ndarray(shape=(1,), dtype=np.float64) for n in range(0)]
    # distance = [np.ndarray(shape=(1,), dtype=np.float64) for n in range(0)]

    # profiles = []
    # coord_x = []
    # coord_y = []
    # distance = []
    # profiles = [np.zeros((1,)) for n in range(0)]
    # coord_x = [np.zeros((1,)) for n in range(0)]
    # coord_y = [np.zeros((1,)) for n in range(0)]
    # distance = [np.zeros((1,)) for n in range(0)]

    # xx = x[x_rotate == x_rotate_values[0]]
    # yy = y[x_rotate == x_rotate_values[0]]
    # ry = y_rotate[x_rotate == x_rotate_values[0]]
    # profile = dem[x_rotate == x_rotate_values[0]]
    # ry_argsort = ry.argsort()
    # profiles = [profile[ry_argsort]]
    # coord_x = [xx[ry_argsort]]
    # coord_y = [yy[ry_argsort]]
    # distance = [np.sort(ry)]

    for rx in x_rotate_values:
        xx = x[x_rotate == rx]
        yy = y[x_rotate == rx]
        ry = y_rotate[x_rotate == rx]
        profile = dem[x_rotate == rx]
        ry_argsort = ry.argsort()
        # profiles.append(profile[ry_argsort])
        # coord_x.append(xx[ry_argsort])
        # coord_y.append(yy[ry_argsort])
        # distance.append(np.sort(ry))

    # return profiles, distance, coord_x, coord_y


@jit(nopython=True)
def dozier(profile):
    """ 1-D Dozier algorithm (see Dozier et al., 1981)

    Without numba: 5.69 ms per loop
    With numba: 14.6 µs per loop
    :param profile:
    :return:
    """
    def slope(obs_point, distant_point):
        if profile[distant_point] <= profile[obs_point]:
            return 0
        else:
            return (profile[distant_point] - profile[obs_point]) / (distant_point - obs_point)

    len_profile = len(profile)
    horizon = np.zeros(len_profile, np.int64)
    horizon[len_profile - 1] = len_profile - 1
    i = len_profile - 2
    while i >= 0:
        j = i + 1
        while "looking for horizon point":
            if slope(i, j) < slope(j, horizon[j]):
                j = horizon[j]
            else:
                if slope(i, j) > slope(j, horizon[j]):
                    horizon[i] = j
                elif slope(i, j) == 0:
                    horizon[i] = i
                else:
                    horizon[i] = horizon[j]
                break
        i -= 1

    return horizon


def dozier_2d(dem, number_of_sectors, distance):
    """

    :param dem:
    :param number_of_sectors:
    :param distance:
    :return:
    """
    pass


def get_horizon(latitude, longitude, dem, ellipsoid=Ellipsoid("WGS84"), distance=0.5, precision=1):
    """ Compute local get_horizon obstruction from Digital Elevation Model

    This function is mainly based on a previous Matlab function
    (see https://fr.mathworks.com/matlabcentral/fileexchange/59421-dem-based-topography-get_horizon-model)
    :param latitude:
    :param longitude:
    :param dem: DigitalElevationModel instance
    :param ellipsoid: Ellipsoid instance
    :param distance: distance in degrees
    :param precision: azimuth precision of resulting horizon in degrees
    :return:
    """

    # Prune DEM and fit to study area
    study_area = dem.get_raster_at(ll_point=(latitude - distance, longitude - distance),
                                   ur_point=(latitude + distance, longitude + distance))

    y_obs, x_obs = study_area.geo_grid.latlon_to_2d_index(latitude, longitude)
    z_obs = study_area.get_value_at(latitude, longitude)

    # Azimuth and elevation
    azimuth = (180/np.pi) * get_azimuth(latitude * np.pi/180, longitude * np.pi/180,
                                        (study_area.geo_grid.latitude - dem.res/2) * np.pi/180,
                                        (study_area.geo_grid.longitude + dem.res/2) * np.pi/180, ellipsoid.e)
    elevation = np.zeros(azimuth.shape)
    elevation[study_area > z_obs] = \
        get_elevation(z_obs, study_area[study_area > z_obs], latitude * np.pi/180,
                      study_area.geo_grid.latitude[study_area > z_obs] * np.pi/180,
                      longitude * np.pi/180, study_area.geo_grid.longitude[study_area > z_obs]
                      * np.pi/180, ellipsoid.e, ellipsoid.a)
    # TODO: understand why "z_obs < study_area" return a numpy ValueError (ambiguous truth value)

    # Elevation vector length
    len_elevation = (90 + precision) // precision

    elevation_dic = dict(ne=np.zeros((y_obs, len_elevation)),
                         e=np.zeros((study_area.x_size - x_obs, 2*len_elevation - 1)),
                         s=np.zeros((study_area.y_size - y_obs, 2*len_elevation - 1)),
                         w=np.zeros((x_obs, 2*len_elevation - 1)), nw=np.zeros((y_obs, len_elevation)))

    azimuth_dic = dict(ne=np.arange(-180, -90 + precision, precision), e=np.arange(-180, 0 + precision, precision),
                       s=np.arange(-90, 90 + precision, precision), w=np.arange(0, 180 + precision, precision),
                       nw=np.arange(90, 180 + precision, precision))

    # Main computation
    # NE & NW
    for n, (az, el) in enumerate(zip(azimuth[:y_obs], elevation[:y_obs])):
        idx_ne = np.digitize(azimuth_dic["ne"], az[x_obs:])
        idx_nw = np.digitize(azimuth_dic['nw'], az[:x_obs])
        elevation_dic["ne"][n, idx_ne < len(az[x_obs:])] = el[x_obs:][idx_ne[idx_ne < len(az[x_obs:])]]
        elevation_dic["nw"][n, idx_nw < len(az[:x_obs])] = el[:x_obs][idx_nw[idx_nw < len(az[:x_obs])]]

    # South
    for n, (az, el) in enumerate(zip(azimuth[y_obs:, ::-1], elevation[y_obs:, ::-1])):
        idx_s = np.digitize(azimuth_dic["s"], az)
        elevation_dic["s"][n, idx_s < len(az)] = el[idx_s[idx_s < len(az)]]

    # East
    for n, (az, el) in enumerate(zip(azimuth[:, x_obs:].transpose(), elevation[:, x_obs:].transpose())):
        idx_e = np.digitize(azimuth_dic["e"], az)
        elevation_dic["e"][n, idx_e < len(az)] = el[idx_e[idx_e < len(az)]]

    # West
    for n, (az, el) in enumerate(zip(azimuth[::-1, :x_obs].transpose(), elevation[::-1, :x_obs].transpose())):
        idx_w = np.digitize(azimuth_dic["w"], az)
        elevation_dic["w"][n, idx_w < len(az)] = el[idx_w[idx_w < len(az)]]

    sun_mask = np.concatenate([elevation_dic[key].max(0) for key in elevation_dic.keys()])
    az_mask = np.concatenate([azimuth_dic[key] for key in azimuth_dic.keys()]) + 180

    horizon = dict(elevation=np.zeros((360 + precision)//precision),
                   azimuth=np.arange(0, 360 + precision, precision))
    for n, az in enumerate(horizon["azimuth"]):
        horizon["elevation"][n] = np.max(sun_mask[az_mask == az])

    horizon["elevation"][-1] = horizon["elevation"][0]

    return horizon


def geo_to_cartesian(latitude, longitude, altitude, e, a):
    """ Convert geographic to cartesian coordinates

    :param latitude:
    :param longitude:
    :param altitude:
    :param e: eccentricity
    :param a: semi-major axis
    :return:
    """
    n = a / np.sqrt(1 - (e**2) * (np.sin(latitude))**2)

    # return X, Y, Z
    return (n + altitude) * np.cos(longitude) * np.cos(latitude), (n + altitude) * np.sin(longitude) * np.cos(
        latitude), (n * (1 - e**2) + altitude) * np.sin(latitude)


def get_azimuth(lat1, lon1, lat2, lon2, e):
    """ Compute azimuth

    :param lat1:
    :param lon1:
    :param lat2:
    :param lon2:
    :param e: eccentricity
    :return:
    """

    # Retrieve isometric latitude
    iso_lat1 = get_isometric_latitude(lat1, e)
    iso_lat2 = get_isometric_latitude(lat2, e)

    # Compute azimuth
    return np.arctan2((lon1 - lon2), (iso_lat1 - iso_lat2))


def get_elevation(h_a, h_b, latitude_a, latitude_b, longitude_a, longitude_b, e, a):
    """ Compute elevation angle

    :param h_a:
    :param h_b:
    :param latitude_a:
    :param latitude_b:
    :param longitude_a:
    :param longitude_b:
    :param e: eccentricity
    :param a: semi-major axis
    :return:
    """
    x_a, y_a, z_a = geo_to_cartesian(latitude_a, longitude_a, h_a, e, a)
    x_b, y_b, z_b = geo_to_cartesian(latitude_b, longitude_b, h_b, e, a)

    inner_prod = (x_b - x_a) * np.cos(longitude_a) * np.cos(latitude_a) + (y_b - y_a) * np.sin(longitude_a) * np.cos(
        latitude_a) + (z_b - z_a) * np.sin(latitude_a)

    # Cartesian norm
    norm = np.sqrt((x_b - x_a) ** 2 + (y_b - y_a) ** 2 + (z_b - z_a) ** 2)

    return np.arcsin(inner_prod/norm) * 180/np.pi


def get_isometric_latitude(latitude, e):
    """ Calculate isometric latitude

    :param latitude:
    :param e: eccentricity
    :return:
    """
    term_1 = np.tan((np.pi/4) + (latitude/2))
    term_2 = ((1 - e * np.sin(latitude)) / (1 + e * np.sin(latitude))) ** (e / 2)

    return np.log(term_1 * term_2)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

# __all__ = []
# __version__ = '0.1'
import os
from tempfile import mkstemp, gettempdir
from utils.check.type import isfile

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2018, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


class LayerTempFile:
    """ Create temp file

    """

    _file_id = 0

    def __init__(self, suffix):
        type(self)._file_id += 1
        self.path = os.path.join(gettempdir(), "tmp_layer_%d%s" % (self._file_id, suffix))
        self._files = [self.path]

    def __enter__(self):
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self

    def __del__(self):
        for file in self._files:
            if isfile(file):
                os.remove(file)


class ShapeTempFile(LayerTempFile):
    """ Create temporary shapefile

    """

    def __init__(self):
        super().__init__(".shp")
        self._files.extend([self.path.replace(".shp", ".shx"), self.path.replace(".shp", ".dbf")])


class GeoJSonTempFile(LayerTempFile):

    def __init__(self):
        super().__init__(".geojson")


class RasterTempFile:
    """ Create temporary raster file

    Class for creating temporary raster files used with gdal-like utilities
    """

    def __init__(self):
        self.path = mkstemp(suffix='.tif')[1]

    def __enter__(self):
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self

    def __del__(self):
        if isfile(self.path):
            os.remove(self.path)



# -*- coding: utf-8 -*-

""" Compute statistics using ZonalStatistics

Compute average from raster cells within a polygon
"""

from gistools.stats import ZonalStatistics
from gistools.raster import RasterMap
from gistools.layer import PolygonLayer

# __all__ = []
# __version__ = '0.1'
__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2017, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


# Create a digital elevation model from SRTM DEM of French Guiana
biomass = RasterMap("biomasse_ressource.tif", no_data_value=-9999)

# Import a geographical layer made of 6 large polygons from a shapefile
layer = PolygonLayer("enp_pn_s_973.shp")

# Convert DEM to layer's coordinate reference system
biomass = biomass.to_crs(layer.crs)

# Create ZonalStatistics instance from biomass and layer. 
zonal_stat = ZonalStatistics(biomass, layer, is_surface_weighted=True, all_touched=True)

# Compute slope average within each polygon
avg = zonal_stat.mean()
std = zonal_stat.std()
print(std)
print(avg)



# -*- coding: utf-8 -*-

""" Tools related to geographic projections

More detailed description.
"""

# import

# __all__ = []
# __version__ = '0.1'
import pyproj

from osgeo import gdal, osr, ogr
from utils.check.type import type_assert
from utils.check.value import check_file

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2018, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


osr.UseExceptions()


def proj4_from_raster(raster_file: str):
    """ Get pyproj Proj4 projection from raster file

    :param raster_file:
    :return:
    """
    check_file(raster_file)

    try:
        proj = gdal.Open(raster_file).GetProjection()
    except RuntimeError:
        raise ValueError("Unable to read raster file '{}'".format(raster_file))

    return proj4_from_wkt(proj)


def proj4_from_layer(layer_file: str):
    """ Get pyproj Proj4 projection from layer file

    :param layer_file:
    :return:
    """
    check_file(layer_file)

    try:
        src_ds = ogr.Open(layer_file)
        return src_ds.GetLayer().GetSpatialRef().ExportToProj4()
    except RuntimeError:
        raise ValueError("Unable to read layer file '%s'" % layer_file)


def proj4_from_wkt(wkt):
    """ Convert wkt srs to proj4

    :param wkt:
    :return:
    """
    srs = osr.SpatialReference()
    srs.ImportFromWkt(wkt)

    return srs.ExportToProj4()


def proj4_from(proj):
    """ Convert projection to proj4 string

    Convert projection string, dictionary, etc.
    to proj4 string
    :param proj:
    :return:
    """
    if type(proj) == int:
        try:
            proj4_str = pyproj.Proj('epsg:%d' % proj).srs
        except (ValueError, RuntimeError):
            raise ValueError("Invalid EPSG code")
    elif type(proj) == str or type(proj) == dict:
        try:
            proj4_str = pyproj.Proj(proj).srs
        except RuntimeError:
            try:
                proj4_str = proj4_from_wkt(proj)
            except (RuntimeError, TypeError):
                raise ValueError("Invalid projection string or dictionary")
    elif type(proj) == pyproj.Proj:
        proj4_str = proj.srs
    else:
        raise ValueError("Invalid projection format: '{}'".format(type(proj)))

    return proj4_str


@type_assert(proj1=(int, str, dict, pyproj.Proj), proj2=(int, str, dict, pyproj.Proj))
def is_equal(proj1, proj2):
    """ Compare 2 projections

    :param proj1:
    :param proj2:
    :return: True or False
    """
    # From an idea from https://github.com/jswhit/pyproj/issues/15
    # Use OGR library to compare projections
    srs = [srs_from(proj1), srs_from(proj2)]

    boolean = (False, True)
    return boolean[srs[0].IsSame(srs[1])]


def ellipsoid_from(proj):
    """ Get Ellipsoid model from projection

    :param proj:
    :return:
    """
    from gistools.coordinates import Ellipsoid

    srs = srs_from(proj)
    ellps = srs.GetAttrValue('Spheroid').replace(' ', '')  # Be sure all spaces in string are removed

    return Ellipsoid(ellps)


def srs_from(proj):
    """ Get spatial reference system from projection

    :param proj:
    :return: SpatialReference instance (osgeo.osr package)
    """
    proj4 = proj4_from(proj)
    srs = osr.SpatialReference()
    srs.ImportFromProj4(proj4)

    return srs


def wkt_from(proj):
    """ Get WKT spatial reference system from projection

    :param proj:
    :return:
    """
    return srs_from(proj).ExportToWkt()



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import math
import os

import pyproj

import numpy as np
import warnings
import geopandas as gpd
from cpc.geogrids import Geogrid
from osgeo import gdal
from rtree import index

from gistools.projections import proj4_from_raster
from gistools.exceptions import GeoGridError, GeoGridWarning

from utils.check.descriptor import protected_property
from utils.check.type import type_assert, check_type, is_iterable
from utils.check.value import check_string, check_file
from utils.sys.reader import read_hdr

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2017, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'
# __all__ = ["Ellipsoid", "GeoGrid", "LocationBase", "Location", "TIME_ZONE"]
# __version__ = '0.1'


REF_ELLIPSOID = pyproj.pj_ellps

TIME_ZONE = {-12, -11, -10, -9.5, -9, -8, -7, -6, -5, -4, -3.5, -3, -2, -1, 0, 1, 2, 3, 3.5, 4,
             4.5, 5, 5.5, 6, 6.5, 7, 8, 8.5, 8.75, 9, 9.5, 10, 10.5, 11, 12, 12.75, 13, 14}

# Be sure numpy return an error when dividing by zero
np.seterr(divide='raise')


class Ellipsoid:
    """ Ellipsoid class instance

    Store parameters of a specified ellipsoid model

    """

    def __init__(self, ellipsoid_model: str):

        ellipsoid_model = check_string(ellipsoid_model, list(REF_ELLIPSOID.keys()))
        self._f = 1 / REF_ELLIPSOID[ellipsoid_model]['rf']
        self._a = REF_ELLIPSOID[ellipsoid_model]['a']
        self._b = self._a - self._a * self._f
        self._e = math.sqrt(1 - (self._b ** 2 / self._a ** 2))
        self._model = ellipsoid_model

    @property
    def f(self):
        return self._f

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b

    @property
    def e(self):
        return self._e

    @property
    def model(self):
        return self._model


class GeoGrid(Geogrid):
    """GeoGrid class instance

    Define regular georeferenced grid based on
    cpc.geogrids module from NOAA
    """
    latitude = protected_property('latitude')
    longitude = protected_property('longitude')
    geo_transform = protected_property('geo_transform')

    def __init__(self, ll_corner, ur_corner, res, geo_type="latlon"):
        """ GeoGrid class constructor

        See super class cpc.geogrids.Geogrid
        :param ll_corner:
        :param ur_corner:
        :param res:
        :param geo_type: grid type ('latlon' or 'equal')
        """
        super().__init__(ll_corner=ll_corner, ur_corner=ur_corner, res=res, type=geo_type)
        if self.lats == [] or self.lons == []:
            raise GeoGridError("GeoGrid has not been rightly defined (empty lats/lons field)")

        # Define geo transform used in raster computing (GDAL syntax)
        self._geo_transform = (self.ll_corner[1] - self.res/2, self.res, 0, self.ur_corner[0] + self.res/2, 0,
                               -self.res)

        # Set lat and lon numpy arrays
        self._lat = np.asarray(self.lats)[::-1]
        self._lon = np.asarray(self.lons)

        # WARNING: sorry guys, but Geogrid implementation of num_x and num_y does not seem robust to me
        # Better to define num_x and num_y as length of lon and lat rather than doing some calculation that fails due
        #  to float precision... (for really small resolutions though)
        self.num_x = len(self._lon)
        self.num_y = len(self._lat)

        # Compute lat/lon meshgrid of pixel centres
        self._longitude, self._latitude = np.meshgrid(self._lon, self._lat)

    @type_assert(lat=(int, float), lon=(int, float))
    def latlon_to_2d_index(self, lat, lon):
        if lat > self._lat.max() + self.res/2 or lat < self._lat.min() - self.res/2 or lon > self._lon.max() + \
                self.res/2 or lon < self._lon.min() - self.res/2:
            raise GeoGridError("Point is out of the geo grid")
        return np.argmin(np.abs(self._lat - lat)), np.argmin(np.abs(self._lon - lon))

    def to_res(self, new_res):
        """ Define geo grid with new resolution

        Return new GeoGrid whose resolution is different
        :param new_res:
        :return:
        """
        new_ll_corner = (self.ll_corner[0] - self.res/2 + new_res/2, self.geo_transform[0] + new_res/2)
        new_ur_corner = (self.geo_transform[3] - new_res/2, self.ur_corner[1] + self.res/2 - new_res/2)

        return GeoGrid(new_ll_corner, new_ur_corner, new_res, geo_type=self.type)

    def __eq__(self, other):

        if other.__class__ == self.__class__:
            return self.ll_corner == other.ll_corner and self.ur_corner == other.ur_corner and self.res == other.res
        else:
            return False

    def __getitem__(self, key):
        """ Get item in GeoGrid instance

        Get item in GeoGrid instance and return GeoGrid (array)
        or latitude and longitude (point)
        :param key:
        :return:
        """

        # If key is only one slice, regard it as the same slice for rows and columns
        if key.__class__ == slice:
            key = (key, key)

        new_res = self.res

        if key.__class__ == tuple:
            if key[0].__class__ == slice or key[1].__class__ == slice:
                try:
                    lat_south, lat_north = self._lat[key[0]][-1], self._lat[key[0]][0]
                except TypeError:
                    lat_south = lat_north = self._lat[key[0]]
                try:
                    lon_west, lon_east = self._lon[key[1]][0], self._lon[key[1]][-1]
                except TypeError:
                    lon_west = lon_east = self._lon[key[1]]
                new_ll_corner, new_ur_corner = (lat_south, lon_west), (lat_north, lon_east)
                return GeoGrid(new_ll_corner, new_ur_corner, new_res, self.type)
            elif key[0].__class__ == int and key[1].__class__ == int:
                return self.latitude.__getitem__(key), self.longitude.__getitem__(key)
            else:
                raise GeoGridError("Invalid indexing")
        elif key.__class__ == int:
            return self.latitude.__getitem__(key), self.longitude.__getitem__(key)
        else:
            raise GeoGridError("Invalid indexing")

    @staticmethod
    def from_geo_file(geo_file: str, res, buffer_accuracy=1, to_crs: str = ''):
        """ Get geo grid from existing geo file (such as shapefile)

        :param geo_file: path to geo file (e.g. shapefile)
        :param res: resolution of the grid
        :param buffer_accuracy: accuracy of the buffer surrounded the region (in degrees)
        :param to_crs: final coordinate reference system of the geo grid (empty string for keeping the original CRS,
        under the form 'epsg:number' or pyproj name otherwise)
        :return: GeoGrid
        """
        # Import locally (to avoid conflicts)
        from gistools.projections import proj4_from

        check_type(geo_file, str, res, (int, float), buffer_accuracy, (int, float), to_crs, str)

        if os.path.isfile(geo_file):
            geo_ds = gpd.read_file(geo_file)
        else:
            raise GeoGridError("{} is not a valid geo file".format(geo_file))

        try:
            to_crs = proj4_from(to_crs)
        except ValueError:
            warnings.warn("Empty or invalid CRS/proj name. Geo grid matching geo file projection", GeoGridWarning)
        else:
            geo_ds = geo_ds.to_crs(to_crs)

        return GeoGrid.from_geopandas(geo_ds, res, buffer_accuracy)

    @staticmethod
    def from_geopandas(geopandas_series, res, buffer_accuracy=1):
        """ Get geo grid from specific geopandas database

        :param geopandas_series:
        :param res:
        :param buffer_accuracy:
        :return:
        """
        check_type(geopandas_series, (gpd.GeoDataFrame, gpd.GeoSeries), res, (int, float), buffer_accuracy, (int,
                                                                                                             float))
        # Default type for geo grid
        geo_type = 'latlon'

        try:
            # If resolution in km/m, type = 'equal'
            if not pyproj.Proj(geopandas_series.crs).crs.is_geographic:
                geo_type = 'equal'
        except AttributeError:
            raise ValueError("Input GeoSeries does not seem to be valid")

        try:
            bounds = geopandas_series.bounds
        except ValueError as e:
            raise GeoGridError("GeoSeries has invalid bounds:\n {}".format(e))

        try:
            ll_corner = (buffer_accuracy * math.floor((bounds.miny.min() + res/2)/buffer_accuracy),
                         buffer_accuracy * math.floor((bounds.minx.min() + res/2)/buffer_accuracy))
            ur_corner = (buffer_accuracy * math.ceil((bounds.maxy.max() - res/2)/buffer_accuracy),
                         buffer_accuracy * math.ceil((bounds.maxx.max() - res/2)/buffer_accuracy))
        except (ValueError, ZeroDivisionError, FloatingPointError):
            raise ValueError("Value of buffer accuracy ({}) is not valid".format(buffer_accuracy))
        else:
            return GeoGrid(ll_corner, ur_corner, res, geo_type)

    @staticmethod
    def from_raster_file(raster_file: str):
        """ Retrieve geo grid from raster file

        :param raster_file: path to raster file
        :return: GeoGrid
        """
        check_type(raster_file, str)

        try:
            source_ds = gdal.Open(raster_file)
        except RuntimeError as error:
            raise error

        geo_transform = source_ds.GetGeoTransform()

        # Warning: ll_corner and ur_corner correspond to grid corners
        # that is pixel centers, but gdal rasters are defined with respect
        # to pixel corners... That's why "+- res/2"
        ll_corner = (geo_transform[3] + source_ds.RasterYSize * geo_transform[5] - geo_transform[5] / 2, geo_transform[
                     0] + geo_transform[1] / 2)
        ur_corner = (geo_transform[3] + geo_transform[5] / 2, geo_transform[0] + source_ds.RasterXSize *
                     geo_transform[1] - geo_transform[1] / 2)

        # Geo type
        if pyproj.Proj(proj4_from_raster(raster_file)).crs.is_geographic:
            geo_type = "latlon"
        else:
            geo_type = "equal"

        if math.isclose(math.fabs(geo_transform[1]), math.fabs(geo_transform[5])):
            geo_grid = GeoGrid(ll_corner, ur_corner, math.fabs(geo_transform[1]), geo_type)
        else:
            geo_grid = None
            warnings.warn('No regular raster: no GeoGrid implemented', Warning)

        return geo_grid

    @staticmethod
    def from_hdr(hdr_file: str):
        """ Get geo grid from envi geospatial header file

        :param hdr_file:
        :return:
        """
        check_file(hdr_file, '.hdr')
        hdr_info = read_hdr(hdr_file)
        if hdr_info['y_res'] != hdr_info['x_res']:
            raise AttributeError("X and Y image resolution must be the same")

        res = hdr_info['y_res']
        ll_corner = (hdr_info['y_origin'] - hdr_info['y_size'] * res + res/2, hdr_info['x_origin'] + res/2)
        ur_corner = (hdr_info['y_origin'] - res/2, hdr_info['x_origin'] + hdr_info['x_size'] * res - res/2)

        if "Lat/Lon" in hdr_info["proj"]:
            geotype = "latlon"
        else:
            geotype = "equal"

        return GeoGrid(ll_corner, ur_corner, res, geotype)


def r_tree_idx(geometry_collection):
    """ Return Rtree spatial index of geometry collection

    :param geometry_collection: geometry collection (list, series, etc.)
    :return:
    """

    idx = index.Index()
    if not is_iterable(geometry_collection):
        geometry_collection = [geometry_collection]

    for i, geom in enumerate(geometry_collection):
        idx.insert(i, geom.bounds)

    return idx
